//
// Copyright (c) 2009 Regents of the University of California. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// 3. The names of its contributors may not be used to endorse or promote
//    products derived from this software without specific prior written
//    permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#include "snn.h"

#define M    100
#define N    1000

int main()
{
	// create a network with N neurons, maximum number of synapes is 100, maximum delay is 20ms, name is global
	CpuSNN s(N, M, 20, "global");

	// create a population or group of neurons with the label "excit", 
	// having 80,000 neurons, and respective izhikevich parameters (a,b,c,d)
	// another function exists where you can specify for each of the 
	// Izhikevich parameters as (mean, std. deviation) pair. This allows random distribution for a,b,c,d parameters
	int g1=s.createInitGroup("excit", N*0.8, 0.02f, 0.2f, -65.0f, 8.0f);

	// create a population or group of neurons with the label "inhib", having 
	// 20,000 neurons and respective Izhikevich parameters(a,b,c,d)
	int g2=s.createInitGroup("inhib", N*0.2, 0.1f,  0.2f, -65.0f, 2.0f);

	// connect from group g2 (inhib) to group g1(excit). The connections are
	// random (uniform distribution) and has M synapses from each neuron in g2 
	// to g1 with initial weight of -5.0. The maximum and minimum delay is 1ms
	s.connect(g2,g1,"random", -5.0f, M, 1, 1);

	// connect from group g1(excit) to group g2(inhib). The connectiona of 
	// random type and have 0.2*M synapses for each neuron in g1 to g2 
	// with initial weight of +6.0. Delay range is 1ms and 20ms.
	s.connect(g1,g2,"random", +6.0f, M*0.2,  1, 20);

	// connect from group g1(excit) to group g1(excit). This is a self connection. 
	// Make 0.8*M synapses randomly to the neurons in g1 with initial weight of +6.0. 	
	// The delay range is randomly generated from a flat distribution between 1ms and 20ms.
	s.connect(g1,g1,"random", +6.0f, M*0.8,  1, 20);

	// define the maximum synaptic weight.
	s.snnProperty("maxWeight",10.0);

	// define the thalamic current into the network. select 0.1% of neurons ( 0.001*N neurons)
	// at random and add a current of 20.0
	s.randomNoiseCurrent(0.1, 20.0);

	// define and set the properties of STDP.
	float ALPHA_LTP = 0.10f;
	float TAU_LTP   = 20.0f;
	float ALPHA_LTD = 0.12f;
	float TAU_LTD   = 20.0f;
	s.stdp( ALPHA_LTP, TAU_LTP, ALPHA_LTD, TAU_LTD);

	// show log every 1 sec (0 to disable logging). You can pass a file pointer or 
	// pass stdout to specify where the log output should go.
	s.setLogCycle(1, stdout);

	// create a spike monitor. This enables storing of the generated
	// spikes for specific population of neurons in group g1
	s.spikeMonitor(g1);

	// create another spike monitor, this time for group g2
	s.spikeMonitor(g2);

	for(int i=0; i < 10; i++) {

		// run the established network for 1 (sec)  and 0 (millisecond), in GPU_MODE
		s.runNetwork(1,0, CPU_MODE);

		// after running the simulation, dump the synaptic weight histogram in a matlab file 
		// with the given name (synhist) and in directory "Histogram".
		// If the second argument is TRUE, then dont overwrite the created 
		// histogram file and for each iteration create a new file with an additional postfix.
		s.printHistogram("Histogram/synhist", true);

		// Dump the firing information in a file Raster/fireInfo1.m. 
		// If second argument is true, dont overwrite the file but 
		// create a new file with a different postfix
		s.rasterPlot(g1, "Raster/fireInfo1.m", true);

		s.rasterPlot(g2, "Raster/fireInfo2.m", true);
	}

	// display the details of the current simulation run
	s.printSimSummary();

}