//
// Copyright (c) 2009 Regents of the University of California. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// 3. The names of its contributors may not be used to endorse or promote
//    products derived from this software without specific prior written
//    permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


  #include "snn.h"

#if ! (_WIN32 || _WIN64)
	#include <string.h>
	#define strcmpi(s1,s2) strcasecmp(s1,s2)
#endif

	// includes, project
	#include <cutil.h>

	MTRand_closed getRandClosed;
	MTRand	      getRand;

	RNG_rand48* gpuRand48;

	/****************************/
	//  debug related stuff.. inspiration from
	// http://www.decompile.com/cpp/faq/file_and_line_error_string.htm

	#define STRINGIFY(x) #x
	#define TOSTRING(x) STRINGIFY(x)
	#define AT __FILE__ ":" TOSTRING(__LINE__)

	inline void error(FILE *fp, const char *location, const char *msg, int sec, int step)
	{
	  fprintf(fp, "(wt=%d,ms=%d) Error at %s: %s\n", sec, step, location, msg);
	}

	inline void debug(FILE *fp, const char *location, const char *msg, int sec, int step)
	{
	  fprintf(fp, "(wt=%d,ms=%d) Executing %s: %s\n", sec, step, location, msg);
	}

	#define DBG(num,fp,loc,msg)  if( DEBUG_LEVEL >= num ) debug(fp,loc,msg,simTimeSec,simTimeMs);

/*********************************************/

	CpuSNN::CpuSNN(int _N, int _M, int _D, char* _name, int _mode): N(_N), M(_M), D(_D)
	{
		fprintf(stdout, "************************************************\n");
		fprintf(stdout, "***** GPU-SNN Simulation Begins Version 0.2 *** \n");
		fprintf(stdout, "************************************************\n");

		neuronInfoSize  = 0;
		synapticInfoSize= 0;
		networkInfoSize = 0;
		spikingInfoSize  = 0;
		addInfoSize     = 0;
		debugInfoSize   = 0;

		fpLog=fopen("tmp_debug.log","w");
		fpProgLog = NULL;
		showLog = 0;	// disable showing log..
		stopHistogram = 0;
		histFileCnt = 0;
		rasterFileCnt = 0;

		currentMode = _mode;
		netInitMode	= _mode;
		memset(&gpu_Ptrs,0,sizeof(network_ptr_t));
		memset(&net_Info,0,sizeof(network_info_t));
		gpu_Ptrs.allocated = false;

		rasterPlotClosed = true;
		rasterFileInfo   = NULL;
		numSpikeMonitor  = 0;

		v	 = new float[N];
		u    = new float[N];
		a	 = new float[N];
		b    = new float[N];
		c	 = new float[N];
		d	 = new float[N];
		I	 = new float[N];

		#ifdef STP
		stpu = new float[N*D];
		stpx = new float[N*D];

		for (int i=0; i < N*D; i++) {
			stpu[i] = STP_U_Exc;
			stpx[i] = 1;
		}
		#endif

		#ifdef CONDUCTANCES
		gAMPA = new float[N];
		gNMDA = new float[N];
		gGABAa = new float[N];
		gGABAb = new float[N];

		for (int i=0; i < N; i++) {
			gAMPA[i] = 0;
			gNMDA[i] = 0;
			gGABAa[i] = 0;
			gGABAb[i] = 0;
		}
		#endif

		lastSpikeTime		= new int[N];
		neuronInfoSize += sizeof(int)*N*8;

		maxWeight	= S_MAX;
		maxWeightChange	= S_MAX;

		Npre = new short[N];
		Npre_exc = new short[N];
		Npost = new short[N];
		cumulativePost = new int[N];
		cumulativePre  = new int[N];
		networkInfoSize += (int)(sizeof(int)*N*3.5);

		for (int i=0; i < MAX_GRP_PER_SNN; i++) {
			grpNeuronType[i]	= UNKNOWN_NEURON;
			grpMaxFiringRate[i] = UNKNOWN_NEURON_MAX_FIRING_RATE;
			grpMonitorId[i]		= -1;
		}

		STDPenabled	 = false;
		simTimeMs		 = 0;	simTimeSec		 = 0;   simTime = 0;
		spikeCountAll1sec = 0;   spikeCountInhibitory1sec  = 0;	spikeCountExcitatory1sec  = 0;
		spikeCountAll     = 0;	spikeCountExcitatory = 0;	spikeCountInhibitory = 0;

		postSynapticIds		= new int[N*M];
		postDelayInfo		= new delay_info_t[N*(D+1)];	// Possible delay values are 0....D (inclusive of D)
		networkInfoSize += (sizeof(int)*N*M)+ (sizeof(delay_info_t)*N*(D+1));

		tmpPreSynapticIds		= new int[N*3*M];				// Temporary array to hold pre-syn connections. will be deleted later
		tmpNeuronType	= new char[N*3*M];				// Temporary array that holds type of connections inhibitory or excitatory
		tmpSynapticDelay	= new uint8_t[N*M];				// Temporary array to store the delays of each connection

		networkName			= new char[sizeof(_name)];
		EtimeTable  = new int[1000+D+1];
		ItimeTable  = new int[1000+D+1];

		for(int p=0; p < (1000+D+1); p++) {
			EtimeTable[p] = 0;
			ItimeTable[p] = 0;
		}

		strcpy(networkName, _name);
		numGrp   = 0;
		numNoise = 0;
		numRandNeurons = 0;

		allocatedN = 0;
		doneReorganization = false;
		char* fname = new char[sizeof(_name)+10];
		sprintf(fname, "%s.dot", _name);
		fpDotty = fopen(fname,"w");

		fprintf(fpDotty, "\
digraph G {\n\
 label = \"%s\";\n\
\tsubgraph wt {\n\
\t\tnode [style=filled];\n\
\t\tlabel = \"%s\";\n\
\t\tcolor=blue;\n", _name, _name);

		if(currentMode == GPU_MODE)
			gpuRand48 = new RNG_rand48(123);

		cutCreateTimer(&timer);
		cutResetTimer(timer);
		cumExecutionTime = 0.0;

	}

	CpuSNN::~CpuSNN()
	{
		try
		{
			printSimSummary(fpLog);
			fclose(fpLog);

			clearRasterData();

			delete[] v;
			delete[] u;
			delete[] a;
			delete[] b;
			delete[] c;
			delete[] d;
			delete[] I;
			delete[] Npre;
			delete[] Npost;
			delete[] lastSpikeTime;
			delete[] postSynapticIds;
			delete[] postDelayInfo;
			delete[] networkName;
			delete[] wt;
			delete[] wtChange;
			delete[] synSpikeTime;
		}

		catch(...)
		{
			fprintf(stderr, "Unknow exception ...\n");
		}
	}

	void CpuSNN::stdp( float _ALPHA_LTP, float _TAU_LTP, float _ALPHA_LTD, float _TAU_LTD)
	{
		assert(_TAU_LTP >= 0.0);
		assert(_TAU_LTD >= 0.0);

		STDPenabled = true;

		ALPHA_LTP = _ALPHA_LTP;
		TAU_LTP   = _TAU_LTP;
		ALPHA_LTD = _ALPHA_LTD;
		TAU_LTD   = _TAU_LTD;

	}

	int CpuSNN::createGroup(char* _name, int _numN)
	{
		int grpId = createGroup( _name, allocatedN, allocatedN + _numN - 1);
		return grpId;
	}

	int CpuSNN::createGroup(char* _name, int _st, int _end)
	{
		assert((_st >=0) && (_st < N));
		assert((_end >=0) && (_end < N));
		assert(_st <= _end);
		int numN = _end - _st + 1;
		assert(numGrp < MAX_GRP_PER_SNN);

		if(allocatedN >= N) {
			printf("Error !!! startingN = %d, endingN = %d, requested Neurons = %d, previously allocated Neurons = %d, total neurons = %d,  \n", _st, _end, numN, allocatedN,  N );
			printf("Error !!! Cannot create the requested group...\n");
			return -1;
		}

		if((allocatedN + numN) > N) {
			printf("Warning !!! Total number of allocated Neurons = %d, Total Neurons = %d, Total Requested Neurons = %d\n", allocatedN,  N, numN);
			printf("Warning !!! Requested neurons in group reduced to %d ...\n", (N-allocatedN));
			numN = N - allocatedN;
		}

		grpStartN[numGrp] = _st;
		grpEndN[numGrp]   = _st + numN - 1;
		grpSizeN[numGrp]  = numN;
		grpName[numGrp]   = new char[strlen(_name)];
		strcpy(grpName[numGrp], _name);
		allocatedN += (_end - _st + 1);

		numGrp++;

		//add a node to the dotty output
		fprintf(fpDotty, "\t\tg%d [label=\"%s\\n(N=%d,D=%d)\"];\n", numGrp-1, _name, grpSizeN[numGrp-1],D);
		fprintf(stdout, "Creating Group %s(id=%d) with N=%d, D=%d\n", _name, numGrp-1, grpSizeN[numGrp-1],D);

		return (numGrp-1);
	}

	// numN => number of neurons in this group
	// _a, _b, _c, _d => Izhikevich parameters
	// numTrace => number of neurons that will be used in firing. mostly first numTrace will be traced
	int  CpuSNN::createInitGroup(char* _name, int _numN, float _a, float _b, float _c, float _d)
	{
		int grpId = createGroup( _name, allocatedN, allocatedN + _numN - 1);
		initGroup(grpId, _a, _b, _c, _d);
		return grpId;
	}

	int CpuSNN::okaySNNRules()
	{
		// ensure that groups do not overlap and thus one neuron lies in only one group
		// ensure that all neurons are covered and we exists in atleast one group

		return true;
	}

	int CpuSNN::createInitGroup(char* _name, int _st, int _end, float _a, float _b, float _c, float _d)
	{
		int grpId = createGroup( _name, _st, _end);
		initGroup(numGrp, _a, _b, _c, _d);
		return grpId;
	}

	// init fixed values for a,b,c and d
	void CpuSNN::initGroup(int groupId, float _a, float _b, float _c, float _d)
	{
		for(int i=grpStartN[groupId]; i <= grpEndN[groupId]; i++) {
			a[i] = _a;
			b[i] = _b;
			c[i] = _c;
			d[i] = _d;
			v[i] = _c;			// initial values for new_v
			u[i] = 0.2f*v[i];   // initial values for u

			lastSpikeTime[i]    = SOME_LARGE_NEGATIVE_VALUE;
			Npre_exc[i]		  = 0;
			Npre[i]			  = 0;
			Npost[i]		  = 0;
			cumulativePost[i] = i*M;
			cumulativePre[i]  = 3*i*M;
		}
	}

	// init variable values for a, b, c, and d
	void CpuSNN::initGroup(int groupId, float _a, float a_sd, float _b, float b_sd, float _c, float c_sd, float _d, float d_sd)
	{
		for(int i=grpStartN[groupId]; i <= grpEndN[groupId]; i++) {
			a[i] = _a + a_sd*(float)getRandClosed();
			b[i] = _b + b_sd*(float)getRandClosed();
			c[i] = _c + c_sd*(float)getRandClosed();
			d[i] = _d + d_sd*(float)getRandClosed();
		}
	}

	// set one specific connection from neuron id 'src' to neuron id 'dest'
	inline void CpuSNN::setConnection(int src, int dest, float initWt, uint8_t dVal)
	{
		assert(dest<=POST_SYN_NEURON_MASK);			// total number of neurons is less than 1 million within a GPU
		assert((dVal >=1) && (dVal <= D));
		assert((initWt < 0)||(initWt > 0));

		// we have exceeded the number of possible connection for one neuron
		if(Npost[src] >= M)		return;

		int p = Npost[src];
		assert((src*M+p) < N*M);

		postSynapticIds[src*M+p]=(Npre[dest]<<POST_SYN_NEURON_BITS)+(dest&POST_SYN_NEURON_MASK); //generate a new postSynapticIds id for the current connection
		tmpPreSynapticIds[dest*3*M+Npre[dest]] = (Npost[src]<<POST_SYN_NEURON_BITS)+(src&POST_SYN_NEURON_MASK);
		tmpNeuronType[dest*3*M + Npre[dest]] = (initWt < 0.0) ? INHIBITORY_NEURON : EXCITATORY_NEURON;

		if(initWt > 0)
			Npre_exc[dest]++;

		tmpSynapticDelay[src*M+p]=dVal;

		Npre[dest]+=1;
		Npost[src]+=1;
	}

	// check if this connection has already been made or not
	inline bool CpuSNN::checkConnectionOkay(int src, int dest)
	{
		if(!(src  < N)) { printf("Src neuron id = %d >  N(%d)\n",  src, N); assert(src < N); }

		// source and destination should be different
		if(src==dest)
			return false;

		// destination should always be less than N
		if(dest>=N)
			return false;

		// dont allow more than one connection
		for(int i=0; i < Npost[src]; i++) {
			if((postSynapticIds[src*M+i]&POST_SYN_NEURON_MASK)==dest)
				return false;
		}

		return true;
	}

	// make 'C' random connections from grpId1 to grpId2
	void CpuSNN::connect_random(int grpId1, int grpId2, float initWt, int C, uint8_t minDelay,  uint8_t maxDelay)
	{
		if(initWt < 0.0 ) {
			if( grpNeuronType[grpId1] == EXCITATORY_NEURON) {
				fprintf(stderr, "Error !!! Group id:%d (%s), has been previously declared to have excitatory connections...\n", grpId1, grpName[grpId1]);
				fprintf(stderr, "now this group is giving out inhibitory connections....Please check the input...\n");
				assert(0);
			}
			grpNeuronType[grpId1] = INHIBITORY_NEURON;
			grpMaxFiringRate[grpId1]   = INHIBITORY_NEURON_MAX_FIRING_RATE;

			if((minDelay!=1)||(maxDelay!=1)) {
				fprintf(stderr, "Allowed delay for inhibitory neuron is only 1 ms ...\n");
				assert((minDelay==1)&&(maxDelay==1));
			}
		}
		else {
			if( grpNeuronType[grpId1] == INHIBITORY_NEURON) {
				fprintf(stderr, "Error !!! Group id:%d (%s), has been previously declared to have inhibitory connections...\n", grpId1, grpName[grpId1]);
				fprintf(stderr, "now this group is giving out excitatory connections....Please check the input...\n");
				assert(0);
			}

			grpNeuronType[grpId1]	 = EXCITATORY_NEURON;
			grpMaxFiringRate[grpId1] = EXCITATORY_NEURON_MAX_FIRING_RATE;
			//printf("Connecting %s(%d) to %s with EXCITATORY connection, id = %d (%d)\n",  grpName[grpId1], grpId1, grpName[grpId2], grpNeuronType[grpId1], EXCITATORY_NEURON);
		}

		grpInitWeights[grpId1]=initWt;	// store temporarily the initialization weight for this group of neuron

		for(int i=grpStartN[grpId1]; i<=grpEndN[grpId1]; i++)
		{
			for(int j=0; j < C; j++)
			{
				bool connOkay=true;
				int dest;
				do {
					dest = grpStartN[grpId2] + (int)((getRand()*grpSizeN[grpId2]));
					connOkay = checkConnectionOkay(i, dest);
				} while(!connOkay);
				uint8_t dVal = minDelay + (int)(0.5+(getRandClosed()*(maxDelay-minDelay)));
				assert((dVal >= minDelay) && (dVal <= maxDelay));
				setConnection(i, dest, initWt, dVal);
			}

		}

		//dotty printf output
		fprintf(fpDotty, "\t\tg%d -> g%d [style=%s, label=\"M=%d\"]\n", grpId1, grpId2, (initWt > 0)?"bold":"dotted", C);
		fprintf(stdout, "Creating Random Connection from Group(%d) to Group(%d) with Probability %f\n",
			             grpId1, grpId2, C*1.0/grpSizeN[grpId2]);
	}

	// print all post-connections...
	void CpuSNN::printPostConnection(FILE *fp)
	{
		for(int i=0; i < numGrp; i++)
			printPostConnection(i,fp);
	}

	// print all the pre-connections...
	void CpuSNN::printPreConnection(FILE *fp)
	{
		for(int i=0; i < numGrp; i++)
			printPreConnection(i,fp);
	}

	// print the connection info of grpId
	void CpuSNN::printPostConnection(int grpId, FILE* fp)
	{
		for(int i=grpStartN[grpId]; i<=grpEndN[grpId]; i++) {
			fprintf(fp, " %3d ( %3d ) : \t", i, Npost[i]);
			// fetch the starting position
			int offset = cumulativePost[i];
			for(int j=0; j < Npost[i]; j++) {
				fprintf(fp, " %3d ( %3d) ", (postSynapticIds[offset+j]&POST_SYN_NEURON_MASK), tmpSynapticDelay[offset+j]);
			}
			fprintf(fp, "\n");
			fprintf(fp, " Delay ( %3d ) : ", i);
			for(int j=0; j < D; j++) {
				fprintf(fp, " %d,%d ", postDelayInfo[i*(D+1)+j].delay_length,
									   postDelayInfo[i*(D+1)+j].delay_index_start);
			}
			fprintf(fp, "\n");
		}
	}

	void CpuSNN::printPreConnection(int grpId, FILE* fp)
	{
		if(!doneReorganization)
		{
			for(int i=grpStartN[grpId]; i<=grpEndN[grpId]; i++) {
				fprintf(fp, " %3d ( %3d ) : \t", i, Npre[i]);
				for(int j=0; j < Npre[i]; j++) {
					fprintf(fp, " %d: %3d (%s) ", j, tmpPreSynapticIds[3*M*i+j]&POST_SYN_NEURON_MASK,
						                     (tmpNeuronType[3*M*i+j]==INHIBITORY_NEURON)?"I":"E");
				}
				fprintf(fp, "\n");
			}
		}

		for(int i=0; i < N; i++)
			fprintf(fp, "Npost[%d] = %d, Npre[%d]=%d, cumPost[%d]=%d, cumPre[%d]=%d\n", i, Npost[i], i, Npre[i], i, cumulativePost[i], i, cumulativePre[i]);
	}

	// make from each neuron in grpId1 to 'M' neurons in grpId2
	void CpuSNN::connect(int grpId1, int grpId2, char* _type, float initWt, int _C, uint8_t minDelay, uint8_t maxDelay)
	{
		assert(grpId1 < numGrp);
		assert(grpId2 < numGrp);
		assert(minDelay <= maxDelay);
		assert(maxDelay <= D);

		if ( strcmpi(_type, "random") == 0)
			connect_random(grpId1, grpId2, initWt, _C, minDelay, maxDelay );
		else
			printf("Invalid connection type( should be 'random', or 'full')\n");

		return;
	}

	// set various parameters for the SNN...
	void CpuSNN::snnProperty(char* _name, float val)
	{
		if(!(strcmp(_name,"maxWeight")))
			maxWeight=val;
		else
			printf("%s: Unknown property set for the SNN\n", _name);
	}

	int CpuSNN::findGrpId(int nid)
	{
		for(int g=0; g < numGrp; g++) {
			if(nid >=grpStartN[g] && (nid <=grpEndN[g]))
				return g;
		}

		fprintf(stderr, "findGrp(): cannot find the group for neuron %d\n", nid);
		assert(0);
	}

	void CpuSNN::updateSpikeMonitor()
	{
		// dont continue if numSpikeMonitor is zero
		if(numSpikeMonitor==0)
			return;

		/* Reset buffer time counter */
		for(int i=0; i < numSpikeMonitor; i++)
			memset(monBufferTimeCnt[i],0,sizeof(int)*1020);

		/* Reset buffer position */
		memset(monBufferPos,0,sizeof(int)*numSpikeMonitor);

		if(currentMode == GPU_MODE) {
			copyFiringInfo_GPU();
		}

		/* Read one spike at a time from the buffer and
		   put the spikes to an appopriate monitor buffer.
		   Later the user may need need to dump these spikes
		   to an output file */
		for(int k=0; k < 2; k++) {
			int* timeTablePtr = (k==0)?EtimeTable:ItimeTable;
			int* fireTablePtr = (k==0)?EfiringTable:IfiringTable;
			for(int t=0; t < 1000; t++) {
				for(int i=timeTablePtr[t+D]; i<timeTablePtr[t+D+1];i++) {
					/* retrieve the neuron id */
					int nid   = fireTablePtr[i];
					//fprintf(fpLog, "%d %d \n", t, nid);

					int grpId = findGrpId(nid);
					int monitorId = grpMonitorId[grpId];
					if(monitorId!= -1) {

						/* retreive the buffers position and group */
						int   pos   = monBufferPos[monitorId]++;
						float frac  = monGrpFrac[monitorId];
						bool  store = false;

						if ( frac >= 0.0f ) {
							/* if frac = 1.0 than always store the spike */
							if ( frac == 1.0f )
								store = true;
							else if (getRandClosed() < frac)
								store = true;
						}

						if (store) {
							monBufferFiring[monitorId][pos] = nid;
							// we store the total firing at time t...
							monBufferTimeCnt[monitorId][t]++;
						}
					} /* if monitoring is enabled for this spike */
				} /* for all spikes happening at time t */
			}  /* for all time t */
		}
	}

	// This function is called every second by simulator...
	// This function updates the firingTable by removing older firing values...
	// and also update the synaptic weights from its derivatives..
	void CpuSNN::updateStateAndFiringTable()
	{
		// CHECK !!!
		// Read the neuron ids that fired in the last D seconds
		// and put it to the beginning of the firing table...
		for(int p=EtimeTable[1000],k=0;p<EtimeTable[1000+D];p++,k++) {
			EfiringTable[k]=EfiringTable[p];
		}

		for(int i=D; i >= 0; i--) {
			EtimeTable[i] = EtimeTable[1000+i]-EtimeTable[1000];
			//ItimeTable[i] = ItimeTable[1000+i]-ItimeTable[1000];
		}

		ItimeTable[D] = 0;

		//update synaptic weights here for all the neurons..
		for(int g=0; g < numGrp; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				int offset = cumulativePre[i];
				for(int j=0; j < Npre_exc[i]; j++) {
					// Excitatory connection changes the synaptic weight
					wt[offset+j] += (0.01f + wtChange[offset+j]);
					wtChange[offset+j]*=0.90f;
					if (wt[offset+j]>maxWeight) wt[offset+j]=maxWeight;
					if (wt[offset+j]<0)  wt[offset+j]=0.0;
				}
			}
		}

		spikeCountAll     += spikeCountAll1sec;
		spikeCountExcitatory += (spikeCountExcitatory1sec-EtimeTable[D]);
		spikeCountInhibitory += spikeCountInhibitory1sec;

		spikeCountInhibitory1sec  = 0;
		spikeCountAll1sec = 0;
		spikeCountExcitatory1sec  = EtimeTable[D];

	}

	void CpuSNN::doECurrentUpdate(int cpu_with_stdp)
	{
		int k = spikeCountExcitatory1sec-1;
		int k_end = EtimeTable[simTimeMs+1];
		int t_pos = simTimeMs;

		while((k>=k_end)&& (k >=0)) {

			// get the neuron id from the index k
			int i  = EfiringTable[k];

			// find the time of firing from the timeTable using index k
			while (!((k >= EtimeTable[t_pos+D])&&(k < EtimeTable[t_pos+D+1]))) {
				t_pos = t_pos - 1;
				assert((t_pos+D-1)>=0);
			}

			// TODO: Instead of using the complex timeTable, can neuronFiringTime value...???
			// Calculate the time difference between time of firing of neuron and the current time...
			int tD = simTimeMs - t_pos;

			assert((tD<D)&&(tD>=0));
			assert(i<N);

			delay_info_t dPar = postDelayInfo[i*(D+1)+tD];

			int offset = cumulativePost[i];

			// for each delay variables
			for(int idx_d = dPar.delay_index_start;
				idx_d < (dPar.delay_index_start + dPar.delay_length);
				idx_d = idx_d+1) {

					// post synaptic info...
					int post_info = postSynapticIds[offset+idx_d];

					// get neuron id
					int p_i = (post_info&POST_SYN_NEURON_MASK);
					assert(p_i<N);

					// get synaptic id
					int s_i = (post_info>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK;
 					assert(s_i<(Npre[p_i]));

					// get the actual position of the synapses and other variables...
					int pos_i = cumulativePre[p_i] + s_i;

					#ifdef CONDUCTANCES
						#ifdef STP
							gAMPA[p_i] += wt[pos_i]*stpx[p_i*D+(simTime-tD)%D]*stpu[p_i*D+(simTime-tD)%D];
							gNMDA[p_i] += wt[pos_i]*stpx[p_i*D+(simTime-tD)%D]*stpu[p_i*D+(simTime-tD)%D];
						#else
							gAMPA[p_i] += wt[pos_i];
							gNMDA[p_i] += wt[pos_i];
						#endif
					#else
						// update the current
						#ifdef STP
							I[p_i] += wt[pos_i]*stpx[p_i*D+(simTime-tD)%D]*stpu[p_i*D+(simTime-tD)%D];
						#else
							I[p_i] += wt[pos_i];
						#endif
					#endif

					// store the firing time for that particular synapses...
					synSpikeTime[pos_i] = simTime;

					// STDP calculation....if required
					if (cpu_with_stdp && STDPenabled) {
						assert(simTime-lastSpikeTime[p_i]>=0);
						int stdp_tDiff = (simTime-lastSpikeTime[p_i]);
						if((stdp_tDiff/TAU_LTD)<25)
							wtChange[pos_i] -= LTD(stdp_tDiff);
					}
			}
			k=k-1;
		}
	}

	void CpuSNN::doICurrentUpdate()
	{
		int k     = spikeCountInhibitory1sec-1;
		int k_end = ItimeTable[simTimeMs+D];

		while((k>=k_end) && (k>=0)) {

			int neuron_id      = IfiringTable[k];
			assert(neuron_id<N);

			delay_info_t dPar = postDelayInfo[neuron_id*(D+1)];

			int offset = cumulativePost[neuron_id];

			// for each connected neuron generate the required inhibitory current..
			for(int idx_d = dPar.delay_index_start;
				idx_d < (dPar.delay_index_start + dPar.delay_length);
				idx_d = idx_d+1) {

					// get synaptic info...
					int post_info = postSynapticIds[offset + idx_d];

					// get neuron id
					int p_i = (post_info&POST_SYN_NEURON_MASK);
					assert(p_i<N);

					// get syn id
					int s_i = (post_info>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK;
					assert(s_i<(Npre[p_i]));

					// get the cumulative position for quick access...
					int pos_i = cumulativePre[p_i] + s_i;

					#ifdef CONDUCTANCES
						#ifdef STP
							gGABAa[p_i] -= wt[pos_i]*stpx[p_i*D+(simTime-1)%D]*stpu[p_i*D+(simTime-1)%D];
							gGABAb[p_i] -= wt[pos_i]*stpx[p_i*D+(simTime-1)%D]*stpu[p_i*D+(simTime-1)%D];
						#else
							gGABAa[p_i] -= wt[pos_i];
							gGABAb[p_i] -= wt[pos_i];
						#endif
					#else
						// update the inhibitory current..
						#ifdef STP
							I[p_i] += wt[pos_i]*stpx[p_i*D+(simTime-1)%D]*stpu[p_i*D+(simTime-1)%D];
						#else
							I[p_i] += wt[pos_i];
						#endif
					#endif

					// store the synpatic firing time for the inhibitory synaptic connection
					// TOD0: since the synaptic weights are fixed... do we need STDP mechanism
					// store the firing time here..???
					synSpikeTime[pos_i] = simTime;
			}
			k=k-1;
		}
	}

	void CpuSNN::initThalInput()
	{
		DBG(3, fpLog, AT, "initThalInput()");

		// reset thalamic currents
		for (int i=0; i < N; i++) I[i] = 0.0;

		// first we do the initial simulation corresponding to the random number generator
		int randCnt=0;
		for (int i=0; i < numNoise; i++)
		{
			 int    noiseGrpId       = noiseGenGroup[i].groupId;
			 float  currentStrength  = noiseGenGroup[i].currentStrength;
			 float  neuronPercentage = noiseGenGroup[i].neuronPercentage;
			 int    ncount = noiseGenGroup[i].ncount;
			 int    nstart = noiseGenGroup[i].nstart;
			 int    nrands = noiseGenGroup[i].rand_ncount;

			 for(int j=0; j < nrands; j++,randCnt++) {
				 int rneuron = nstart + (int)(ncount*getRand());
				 randNeuronId[randCnt]= rneuron;
				 //fprintf(fpLog, " %d %d \n", simTimeSec*1000+simTimeMs, randNeuronId[randCnt]);
				 // assuming there is only one driver at a time.
				 // More than one noise input is not correct...
				 I[rneuron] = currentStrength;
			 }
		}
	}

	void CpuSNN::dumpSpikeBuffToFile(FILE *fpSpike, int gid)
	{
		DBG(2, fpLog, AT, "dumpSpikeBuffToFile()");

		assert((gid >= 0)&& (gid < numGrp));
		assert(fpSpike != NULL);

		int  monId = grpMonitorId[gid];
		if(monId == -1) {
			fprintf(stderr, "RasterPlot(gid=%d) failed... Group %s not monitored..\n", gid, grpName[gid]);
			fprintf(stderr, "call spikeMonitor(gid=%d) function in your code to ensure monitoring of group %s\n", gid, grpName[gid]);
			return;
		}

		int* fBuf  = monBufferFiring[monId];
		int* tBuf  = monBufferTimeCnt[monId];
		int pos    = 0;
		for (int t=0; t < 1000; t++) {
			for(int i=0; i<tBuf[t];i++,pos++) {
				fprintf(fpSpike, "%d %d\n", t, fBuf[pos]);
			}
		}

		rasterPlotClosed = false;

	}

	void CpuSNN::saveFiringInfo(char* fname, FILE *fpSpike)
	{
		DBG(2, fpLog, AT, "saveFiringInfo()");

		char fnname[100];
		sprintf(fnname,"%s.txt",fname);
		FILE *fp_save=fopen(fnname,"w");
		for(int i=0; i < 1000+D+1;i++) {
			fprintf(fp_save, "(%d,%d): %d, %d\n", simTimeSec, i, EtimeTable[i], ItimeTable[i]);
		}
		fclose(fp_save);

		FILE*fpDat;
		if(fpSpike!=NULL) {
			fpDat = fpSpike;
		}
		else {
			sprintf(fnname,"%s_sec%d.m", fname, simTimeSec);
			fpDat=fopen(fnname, "w");
			fprintf(fpDat,"N=%d;Ne=%d,Ni=%d,M=%d;simsec=%d;\n", N, Ne, Ni, M, simTimeSec);
			fprintf(fpDat, "fireInfo = [\n");
		}

		#define MAX_SAVING_SIZE_E	800
		#define MAX_SAVING_SIZE_I	200

		// first version, we randomly pick 800 neurons and store their trace..
		int cnt=0;
		for(int t=0; t < 1000; t++) {
			for(int i=EtimeTable[t+D]; i<EtimeTable[t+D+1];i++) {
					if(rand() < (MAX_SAVING_SIZE_E/N))
						fprintf(fpDat, "%d %d\n", t, cnt++);
			}
		}

		// first version, we randomly pick 200 inhibitory neuron and store their trace...
		for(int t=0; t < 1000; t++) {
			for(int i=ItimeTable[t+D]; i<ItimeTable[t+D+1];i++) {
				if(rand() < (MAX_SAVING_SIZE_I/N))
					fprintf(fpDat, "%d %d\n", t, cnt++);
			}
		}

		if(fpSpike==NULL) {
			fprintf(fpDat, "];\n");
			fclose(fpDat);
		}
	}

	// monitor spikes at gid. Its enough to monitor fracMonitor fraction
	// of neurons. We are hoping the maximum rate is 25Hz.
	// spikes that exceed this size limit is dropped...
	void CpuSNN::spikeMonitor(int gid, float fracMonitor, int maxRate)
	{
		DBG(2, fpLog, AT, "spikeMonitor() Added");

		// store the gid for further reference
		monGrpId[numSpikeMonitor]	= gid;

		// also inform the grp that it is being monitored...
		grpMonitorId[gid]			= numSpikeMonitor;

		// note down the fraction that need to be monitored..
		monGrpFrac[numSpikeMonitor] = fracMonitor;

		// if maxRate is not specified it will default to the group max rate
		if(maxRate==-1) {
			maxRate	= grpMaxFiringRate[gid];
		}

		// count the size of the buffer for storing 1 sec worth of info..
		// only the last second is stored in this buffer...
		int buffSize = (int)(maxRate*fracMonitor*grpSizeN[gid])+10;

		// store the size for future comparison etc.
		monBufferSize[numSpikeMonitor] = buffSize;

		// reset the position of the buffer pointer..
		monBufferPos[numSpikeMonitor]  = 0;

		// create the new buffer for keeping track of all the spikes in the system
		monBufferFiring[numSpikeMonitor] = new int[buffSize];
		monBufferTimeCnt[numSpikeMonitor]= new int[1020];
		memset(monBufferTimeCnt[numSpikeMonitor],0,sizeof(int)*1020);

		numSpikeMonitor++;

		// oh. finally update the size info that will be useful to see
		// how much memory are we eating...
		debugInfoSize += sizeof(int)*buffSize + sizeof(int)*1020;

	}

	void CpuSNN::updateRasterPlot()
	{
		DBG(2, fpLog, AT, "updateRasterPlot()");

		if(!rasterPlotClosed) {
			fileInfo_t* wt=rasterFileInfo;
			while(wt) {
				if(wt->fp!=NULL) {
					fprintf(wt->fp, " ] ; %% ends firing Info \n");
					fclose(wt->fp);
					wt->fp=NULL;
					wt = wt->next;
				}
			}
			rasterPlotClosed = true;
			rasterFileCnt++;
		}
	}

	/* close files and also clear all data used
	   by raster generation function */
	void CpuSNN::clearRasterData()
	{
		DBG(2, fpLog, AT, "clearRasterData()");
		updateRasterPlot();
		fileInfo_t* wt=rasterFileInfo;
		while(wt) {
			fileInfo_t* tmp = wt->next;
			//free(wt->fname);
			delete wt;
			wt = tmp;
		}
	}

	void appendFileName(char* newfname, char* fname, int val)
	{
		assert(newfname!=NULL);
		assert(fname!=NULL);

		char* extStr=strchr(fname,(int) '.');
		char* defaultExt=".m";
		size_t len;
		// no extension... so use default extension
		if(extStr==NULL)  {
			extStr = defaultExt;
			len = strlen(fname);
		}
		else {
			// length of the file name with extension
			len =(strlen(fname)-strlen(extStr));
		}

		strcpy(newfname, fname);

		//file name without extension..
		newfname[len]='\0';

		//create a new file name if we should not overwrite..
		sprintf(newfname, "%s_%d%s", newfname, val, extStr);
		return;
	}

	void CpuSNN::rasterPlot(int gid, char* fname, bool dontOverWrite)
	{
		DBG(2, fpLog, AT, "rasterPlot() called");

		bool fileExist     = false;
		bool fileNameExist = false;
		fileInfo_t* wt=rasterFileInfo;

		/* scan for the specific file name in our list */
		while(wt != NULL) {
			if(strcmpi(fname, wt->fname)==0) {
				fileNameExist = true;
				if(wt->fp!=NULL && !dontOverWrite)
					fileExist = true;
				break;
			}
			wt=wt->next;
		}

		/* file does not exists */
		if(!fileNameExist) {
			if(rasterFileInfo==NULL) {
				/* first time rasterFileInfo is created */
				rasterFileInfo = new fileInfo_t;
				rasterFileInfo->next = NULL;
			}
			else {
				fileInfo_t* node = new fileInfo_t;
				node->next = rasterFileInfo;
				rasterFileInfo = node;
			}
			/* Create the file and store the file name in the list */
			rasterFileInfo->fname = strdup(fname);
			wt = rasterFileInfo;
		}

		if(!fileExist)	{
			char newfname[100];
			if(strlen(fname)>100) {
				DBG(0,stderr,AT,"file name size is larger then available local memory (100 characters)..\nInput file name is -->");
				DBG(0, stderr, AT, newfname);
				wt->fp=NULL;
				return;
			}
			if(dontOverWrite)
				appendFileName(newfname, fname, rasterFileCnt);
			else
				strcpy(newfname, fname);

			wt->fp    = fopen(newfname,"w");

			if(wt->fp==NULL) {
				fprintf(stderr, "Unable to open raster log file --> %s...\n", newfname);
				return;
			}

			fprintf(wt->fp, "%%%% firing info for <<%s>> with gid=%d\n", grpName[gid], gid);
			fprintf(wt->fp, "firing%d = [\n", gid);
		}

		dumpSpikeBuffToFile(wt->fp, gid);

	}

	void  CpuSNN::globalStateUpdate()
	{
		// now we update the state of all the neurons
		for(int g=0; g < numGrp; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				#ifdef CONDUCTANCES
				for (int j=0; j<COND_INTEGRATION_SCALE; j++) {
					float NMDAtmp = (v[i]+80)*(v[i]+80)/60/60;
					I[i] -= (gAMPA[i]*(v[i]-0) + gNMDA[i]*NMDAtmp/(1+NMDAtmp)*(v[i]-0) + gGABAa[i]*(v[i]+70) + gGABAb[i]*(v[i]+90))/COND_INTEGRATION_SCALE;

					v[i]+=((0.04f*v[i]+5)*v[i]+140-u[i]+I[i])/COND_INTEGRATION_SCALE;
					if (v[i] > 30) v[i] = 30;
					u[i]+=a[i]*(0.2f*v[i]-u[i])/COND_INTEGRATION_SCALE;
				}
				#else
				v[i]+=0.5f*((0.04f*v[i]+5)*v[i]+140-u[i]+I[i]); // for numerical stability
				v[i]+=0.5f*((0.04f*v[i]+5)*v[i]+140-u[i]+I[i]); // time step is 0.5 ms
				if (v[i] > 30) v[i] = 30;
				u[i]+=a[i]*(0.2f*v[i]-u[i]);
				#endif
			}
		}
	}


	// show the status of the simulator...
	// when onlyCnt is set, we print the actual count instead of frequency of firing
	void CpuSNN::showStatus(int simType, int onlyCnt)
	{
		if(simType == GPU_MODE) {
			int gpu_secIfireCnt, gpu_secEfireCnt;
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &gpu_secEfireCnt, "secEfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &gpu_secIfireCnt, "secIfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			spikeCountAll1sec = gpu_secIfireCnt + gpu_secEfireCnt;
			spikeCountInhibitory1sec  = gpu_secIfireCnt;
		}
		if(onlyCnt) {
			fprintf(fpLog, "(t=%d) Firing Count = %d , (excit=%d , inhib=%d )",
				          simTimeSec, spikeCountAll1sec, (spikeCountAll1sec-spikeCountInhibitory1sec), spikeCountInhibitory1sec);
		    if(fpProgLog)
			   fprintf(fpProgLog, "(t=%d) Firing Count = %d , (excit=%d, inhib=%d)",
			              simTimeSec, spikeCountAll1sec, (spikeCountAll1sec-spikeCountInhibitory1sec), spikeCountInhibitory1sec);
		}
		else {
		    fprintf(fpLog, "(t=%d) Firing Rate = %3.3f Hz, (excit=%3.3f Hz, inhib=%3.3f Hz)",
				   simTimeSec, spikeCountAll1sec*1.0f/N, (spikeCountAll1sec-spikeCountInhibitory1sec)*1.0f/Ne, spikeCountInhibitory1sec*1.0f/Ni);
		    if(fpProgLog)
			   fprintf(fpProgLog, "(t=%d) Firing Rate = %3.3f Hz, (excit=%3.3f Hz, inhib=%3.3f Hz)",
			       simTimeSec, spikeCountAll1sec*1.0f/N, (spikeCountAll1sec-spikeCountInhibitory1sec)*1.0f/Ne, spikeCountInhibitory1sec*1.0f/Ni);
		}
	}

	void CpuSNN::startCPUTiming()
	{
		prevCpuExecutionTime = cumExecutionTime;
	}

	void CpuSNN::resetCPUTiming()
	{
		prevCpuExecutionTime = cumExecutionTime;
		cpuExecutionTime     = 0.0;
	}

	void CpuSNN::stopCPUTiming()
	{
		cpuExecutionTime += (cumExecutionTime - prevCpuExecutionTime);
	}

	void CpuSNN::startGPUTiming()
	{
		prevGpuExecutionTime = cumExecutionTime;
	}

	void CpuSNN::resetGPUTiming()
	{
		prevGpuExecutionTime = cumExecutionTime;
		gpuExecutionTime     = 0.0;
	}

	void CpuSNN::stopGPUTiming()
	{
		gpuExecutionTime += (cumExecutionTime - prevGpuExecutionTime);
	}

	// Run the simulation for n sec
	void CpuSNN::runNetwork(int _nsec, int _nmsec, int simType)
	{
		DBG(2, fpLog, AT, "runNetwork() called");

		assert(_nmsec >= 0);
		assert(_nsec  >= 0);
		assert(simType == CPU_MODE || simType == GPU_MODE);

		// First time when the network is run we do various kind of space compression,
		// and data structure optimization to improve performance and save memory.
		if(!doneReorganization)
			reorganizeNetwork();

		if((simType == GPU_MODE) && (gpu_Ptrs.allocated == false))
			allocate();

		currentMode = simType;

		cutResetTimer(timer);
		cutStartTimer(timer);

		// if nsec=0, simTimeMs=10, we need to run the simulator for 10 timeStep;
		// if nsec=1, simTimeMs=10, we need to run the simulator for 1*1000+10, time Step;
		for(int i=0; i < (_nsec*1000 + _nmsec); i++) {

			initThalInput();

			if(simType == CPU_MODE)
				doSnnSim(1);
			else
				doGPUSim(1);

			// done one second worth of simulation
			// update relevant parameters...now
			if(++simTimeMs == 1000) {
				if(showLog)
					if(showLogCycle==showLog++) {
						showStatus(currentMode);
						showLog=1;
					}

				updateSpikeMonitor();

				updateRasterPlot();

				if(simType == CPU_MODE)
					updateStateAndFiringTable();
				else
					updateStateAndFiringTable_GPU();

				simTimeMs = 0;
				simTimeSec++;
			}

			simTime++;

		}

		// keep track of simulation time...
		cutStopTimer(timer);
		lastExecutionTime = cutGetTimerValue(timer);
		fprintf(fpLog, "(t%s = %2.2f sec)\n",  (currentMode == GPU_MODE)?"GPU":"CPU", lastExecutionTime/1000);
		fprintf(stdout, "(t%s = %2.2f sec)\n", (currentMode == GPU_MODE)?"GPU":"CPU", lastExecutionTime/1000);
		cumExecutionTime += lastExecutionTime;

	}

	// This would supply the required random current to the specific percentage of neurons. If groupId=-1,
	// then the neuron is picked from the full population.
	// If the groupId=n, then only neurons are picked from selected group n.
	void CpuSNN::randomNoiseCurrent(float neuronPercentage, float currentStrength, int groupId)
	{
		DBG(2, fpLog, AT, "randomNoiseCurrent() added");

		int noiseGrpId = groupId + 1;
		if(noiseGrpId==0)
			fprintf(fpDotty, "\t\tr%d [shape=box, label=\"Global Random Noise\\n(I=%2.2fmA, frac=%2.2f%%)\"];\n", numNoise, currentStrength, neuronPercentage);
		else
			fprintf(fpDotty, "\t\tr%d [shape=box, label=\"Random Noise\\n(I=%2.2fmA, frac=%2.2f%%)\"];\n", numNoise, currentStrength, neuronPercentage);

		int  numNeuron  = ((groupId==-1)? N: grpSizeN[groupId]);
		int  nstart     = ((groupId==-1)? 0: grpStartN[groupId]);
		noiseGenGroup[numNoise].groupId			 = groupId;
		noiseGenGroup[numNoise].currentStrength  = currentStrength;
		noiseGenGroup[numNoise].neuronPercentage = neuronPercentage;
		noiseGenGroup[numNoise].ncount			 = numNeuron;
		noiseGenGroup[numNoise].nstart			 = nstart;
		noiseGenGroup[numNoise].rand_ncount		 = (int)(numNeuron*(neuronPercentage/100.0));
		if(noiseGenGroup[numNoise].rand_ncount<=0)
			noiseGenGroup[numNoise].rand_ncount = 1;

		if(noiseGrpId != 0)
			fprintf(fpDotty, "\t\tr%d -> g%d [label=\" n=%d \"];\n", numNoise, groupId, (int) (numNeuron*(neuronPercentage/100.0)));

		numNoise++;
	}

	void CpuSNN::printHistogram(char* histName, bool dontOverWrite, unsigned int histSize)
	{
		DBG(2, fpLog, AT, "printHistogram()");

		if(histName==NULL)
			putHistogram(fpLog, simTimeSec, "synhist", histSize, dontOverWrite);
		else
			putHistogram(fpLog, simTimeSec, histName, histSize, dontOverWrite);
	}

	void CpuSNN::printSimSummary(FILE *fp)
	{
		DBG(2, fpLog, AT, "printSimSummary()");
		float etime;
		if(currentMode == GPU_MODE)	 {
			stopGPUTiming();
			etime = gpuExecutionTime;
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &spikeCountExcitatory, "secEfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &spikeCountInhibitory, "secIfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			spikeCountAll1sec = spikeCountInhibitory + spikeCountExcitatory;
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &spikeCountExcitatory, "totEfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &spikeCountInhibitory, "totIfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
			spikeCountAll      = spikeCountInhibitory + spikeCountExcitatory;
		}
		else {
			stopCPUTiming();
			etime = cpuExecutionTime;
		}

		fprintf(fp, "\n*** Network configuration dumped in %s.dot file...\n\
Use graphViz to see the network connectivity...\n\n", networkName);
		fprintf(fp, "*********** %s Simulation Summary **********\n", (currentMode == GPU_MODE)?"GPU":"CPU");
		fprintf(fp, "Network Parameters: \n\tN = %d (Ne:Ni=%2.1f:%2.1f), M = %d, D = %d\n", N, 100.0*Ne/N, 100.0*Ni/N, M, D);
		fprintf(fp, "Timing: \n\tModel Simulation Time = %d sec \n\tActual Execution Time = %4.2f sec\n",  simTimeSec, etime/1000.0);
		fprintf(fp, "Average Firing Rate \n\tExcitatory = %3.3f Hz \n\tInhibitory = %3.3f Hz \n\tOverall = %3.3f Hz\n",
			spikeCountExcitatory/(1.0*simTimeSec*Ne), spikeCountInhibitory/(1.0*simTimeSec*Ni), spikeCountAll/(1.0*simTimeSec*N));
		fprintf(fp, "Overall Firing Count: \n\tExcitatory = %d \n\tInhibitory = %d \n\tTotal = %d\n",
			spikeCountExcitatory, spikeCountInhibitory, spikeCountAll );
		fprintf(fp, "**************************************\n\n");
	}

	void CpuSNN::resetNetwork()
	{
		// Reset v,u,firing time values to default values...
		// Reset wt,wtChange,pre-firingtime values to default values...
		// Reset firing tables and time tables to default values..
	}

	void CpuSNN::doSnnSim(int cpu_with_stdp)
	{
		int signaled100Hz = 0;
		for(int g=0; (g < numGrp) & !signaled100Hz; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {

				#ifdef STP
				int ind = i*D+simTime%D;
				int ind_1 = i*D+(simTime-1)%D;
				if(grpNeuronType[g] == EXCITATORY_NEURON) {
					stpx[ind] = stpx[ind_1] + (1-stpx[ind_1])/STP_tD_Exc;
					stpu[ind] = stpu[ind_1] + (STP_U_Exc - stpu[ind_1])/STP_tF_Exc;
				} else {
					stpx[ind] = stpx[ind_1] + (1-stpx[ind_1])/STP_tD_Inh;
					stpu[ind] = stpu[ind_1] + (STP_U_Inh - stpu[ind_1])/STP_tF_Inh;
				}
				#endif
				#ifdef CONDUCTANCES
				gAMPA[i] -= gAMPA[i]/COND_tAMPA;
				gNMDA[i] -= gNMDA[i]/COND_tNMDA;
				gGABAa[i] -= gGABAa[i]/COND_tGABAa;
				gGABAb[i] -= gGABAb[i]/COND_tGABAb;
				#endif

				if (v[i] >= 30.0) {

					v[i] = -65.0;
					u[i] += d[i];
					lastSpikeTime[i] = simTime;

					#ifdef STP
					stpx[ind] = stpx[ind] - stpu[ind_1]*stpx[ind_1];
					if(grpNeuronType[g] == EXCITATORY_NEURON)
						stpu[ind] = stpu[ind] + STP_U_Exc*(1-stpu[ind_1]);
					else
						stpu[ind] = stpu[ind] + STP_U_Inh*(1-stpu[ind_1]);
					#endif

					if(grpNeuronType[g] == EXCITATORY_NEURON) {
						EfiringTable[spikeCountExcitatory1sec] = i;
						spikeCountExcitatory1sec++;

						if(spikeCountExcitatory1sec >= maxSpikesExcitatory) {
							if (!signaled100Hz)
								printf("Average excitatory spike rate >%dhz for this ms; additional spikes are lost.\n",EXCITATORY_NEURON_MAX_FIRING_RATE);
							signaled100Hz = 1;
							spikeCountExcitatory1sec = maxSpikesExcitatory-1;
							break;
						}
					}
					else {
						IfiringTable[spikeCountInhibitory1sec] = i;
						spikeCountInhibitory1sec++;

						if(spikeCountInhibitory1sec >= maxSpikesInhibitory) {
							if (!signaled100Hz)
								printf("Average inhibitory spike rate >%dhz for this ms; additional spikes are lost.\n",INHIBITORY_NEURON_MAX_FIRING_RATE);
							signaled100Hz = 1;
							spikeCountInhibitory1sec = maxSpikesInhibitory-1;
							break;
						}
					}

					if (cpu_with_stdp && STDPenabled) {
						int pos_ij = cumulativePre[i];
						for(int j=0; j < Npre_exc[i]; pos_ij++, j++) {
							if(!(simTime-synSpikeTime[pos_ij]>=0)) {
								fprintf(stderr, "simTime=%d, pre_firing=%d\n", simTime, synSpikeTime[pos_ij]);
								assert(0);
							}
							// dont do LTP if time difference is a lot..
							if((simTime-synSpikeTime[pos_ij])/TAU_LTP<25)
								wtChange[pos_ij] += LTP(simTime-synSpikeTime[pos_ij]);
						}
					}
					spikeCountAll1sec++;
				}
			}
		}

		EtimeTable[simTimeMs+D+1] = spikeCountExcitatory1sec;
		ItimeTable[simTimeMs+D+1] = spikeCountInhibitory1sec;

		doECurrentUpdate(cpu_with_stdp);
		doICurrentUpdate();

		globalStateUpdate();

		return;

	}

	// The post synaptic connections are sorted based on delay here so that we can reduce storage requirement
	// and generation of spike at the post-synaptic side.
	// We also create the delay_info array has the delay_start and delay_length parameter
	void CpuSNN::reorganizeDelay()
	{
		for(int g=0; g < numGrp; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				int jPos=0;					// this points to the top of the delay queue
				int cumN=cumulativePost[i];
				int cumDelayStart=0;
				for(int td = 0; td < D; td++) {
					int j=jPos;				// start searching from top of the queue until the end
					int cnt=0;				// store the number of nodes with a delay of td;
					while(j < Npost[i]) {
						// found a node j with delay=td and we put
						// the delay value = 1 at array location td=0;
						if(td==(tmpSynapticDelay[cumN+j]-1)) {

							assert(jPos<Npost[i]);

							// Put the node j to the top of the delay queue
							int tmp = postSynapticIds[cumN+j];
							postSynapticIds[cumN+j]=postSynapticIds[cumN+jPos];
							postSynapticIds[cumN+jPos]=tmp;

							// Ensure that you have shifted the delay accordingly....
							uint8_t tmp_delay = tmpSynapticDelay[cumN+j];
							tmpSynapticDelay[cumN+j]    = tmpSynapticDelay[cumN+jPos];
							tmpSynapticDelay[cumN+jPos] = tmp_delay;

							jPos=jPos+1;
							cnt=cnt+1;
						}
						j=j+1;
					}

					// update the delay_length and start values...
					postDelayInfo[i*(D+1)+td].delay_length		= cnt;
					postDelayInfo[i*(D+1)+td].delay_index_start  = cumDelayStart;
					cumDelayStart += cnt;

					assert(cumDelayStart <= Npost[i]);
				}
				//total cumulative delay should be equal to number of post-synaptic connections	at the end of the loop
				assert(cumDelayStart == Npost[i]);
				for( int j=1; j < Npost[i]; j++) {
					int cumN=cumulativePost[i];
					if( tmpSynapticDelay[cumN+j] < tmpSynapticDelay[cumN+j-1]) {
						fprintf(stderr, "Post-synaptic delays not sorted corrected...\n");
						fprintf(stderr, "id=%d, delay[%d]=%d, delay[%d]=%d\n",
							i, j, tmpSynapticDelay[cumN+j], j-1, tmpSynapticDelay[cumN+j-1]);
						assert( tmpSynapticDelay[cumN+j] >= tmpSynapticDelay[cumN+j-1]);
					}
				}
			}
		}
	}

	/* we calculate global properties like how many random number need to be generated
	each cycle, etc */
	void CpuSNN::updateRandomProperty()
	{
		numRandNeurons = 0;
		for (int i=0; i < numNoise; i++) {
			 numRandNeurons += noiseGenGroup[i].rand_ncount;
		}

		assert(numRandNeurons>=0);

		randNeuronId = new int[numRandNeurons];

		addInfoSize += (sizeof(int)*numRandNeurons);
	}

	char* extractFileName(char *fname)
	{
		char* varname = strrchr(fname, '\\');
		size_t len1 = strlen(varname+1);
		char* extname = strchr(varname, '.');
		size_t len2 = strlen(extname);
		varname[len1-len2]='\0';
		return (varname+1);
	}

	void CpuSNN::dumpHistogram(int* table_sHist, int* table_sdHist, FILE *fph, int sec, char *fname, int histSize)
	{
		DBG(2, fpLog, AT, "dumpHistogram");

		fprintf(fph,"N=%d;simTimeSec=%d;M=%d;\n", N,sec,M);
		fprintf(fph, "synhist_s=[");
		for(int i=0; i < (histSize); i++) {
			fprintf(fph, "%d ", table_sHist[i]);
		}
		fprintf(fph, "];\n");
		fprintf(fph, "synhist_sd=[");
		for(int i=0; i < (histSize); i++) {
			fprintf(fph, "%d ", table_sdHist[i]);
		}

		fprintf(fph, "];");
		fflush(fph);
		fclose(fph);
	}

	void CpuSNN::generateHistogram(int* sHist, int* sdHist, int histSize, FILE* fp)
	{
		assert(histSize >=0 );
		memset(sHist,0,sizeof(int)*(histSize+2));
		memset(sdHist,0,sizeof(int)*(histSize+2));

		float sStep  = (maxWeight)/(histSize);
		float sdStep = maxWeightChange/(histSize);

		float *t_s  = wt;
		float *t_sd = wtChange;
		for(int g=0; g < numGrp; g++) {
			//plot only excitatory neuron'wt weight changes...
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				int pos = cumulativePre[i];
				for(int j=0; j < Npre_exc[i]; j++, pos++) {
					unsigned int histPos_s,histPos_sd;
					float tmp_s  = t_s[pos];
					float tmp_sd = t_sd[pos];

					histPos_s  = (unsigned int) ((tmp_s/sStep));
					histPos_sd = (unsigned int) (((tmp_sd+(maxWeightChange/2))/sdStep));

					if (histPos_s  >= histSize)
						sHist[histSize+1]++;	//overflows...
					else if(histPos_s  < 0)
						sHist[0]++;				//underflows...
					else
						sHist[histPos_s+1]++;	//within range

					if(histPos_sd >= histSize)
						sdHist[histSize+1]++;	//overflows in wtChange...
					else if(histPos_sd < 0)
						sdHist[0]++;				//underflows in wtChange...
					else
						sdHist[histPos_sd+1]++;		//within range

					if((histPos_s>(histSize))  ||
					   (histPos_sd>(histSize)) ||
					   (histPos_sd < 0) ||
					   (histPos_s  < 0)) {
						   if(fp!=NULL) {
							   fprintf(fp,"CPU HISTOGRAM Calculation error\n");
							   fprintf(fp,"wt=%f,histS=%d,wtChange=%f,histSd=%d\n", tmp_s, histPos_s, tmp_sd, histPos_sd);
							   fprintf(fp,"i=%d,j=%d,Npre=%d,pos=%d\n", i,j,Npre[i],pos);
						   }
					}
				}
			}
		}
	}

	void CpuSNN::putHistogram(FILE *fp, int sec, char* fname, unsigned int histSize, bool dontOverWrite)
	{
		DBG(2, fpLog, AT, "cpu_getHistogram");

		// gpu mode currently support fixed size histogram only...
		if (currentMode == GPU_MODE) {
			histSize = HISTOGRAM_SIZE;
		}

		int* cpu_sHist  = new int[2*(histSize+2)];

		if(currentMode == CPU_MODE)
			generateHistogram(&cpu_sHist[0] , &cpu_sHist[histSize+2], histSize, fp);
		else
			generateHistogram_GPU(&cpu_sHist[0], &cpu_sHist[histSize+2], histSize);

		char* newfname;
		newfname = (char*) malloc(strlen(fname)+100);

		if(dontOverWrite)
			appendFileName(newfname, fname, histFileCnt++);
		else
			strcpy(newfname, fname);

		FILE *fph=fopen(newfname,"w");
		if(fph==NULL) {
			fprintf(stderr, "Histogram file <<%s>> cannot be opened\n", newfname);
			free(newfname);

			delete[] cpu_sHist;
			return;
		}

		DBG(2, fpLog, AT, newfname);
		dumpHistogram(&cpu_sHist[0], &cpu_sHist[histSize+2], fph, sec, newfname, histSize);

		free(newfname);
		delete[] cpu_sHist;

	}

#define COMPACTION_ALIGNMENT_PRE  16
#define COMPACTION_ALIGNMENT_POST 0

#define SETPOST_INFO(name, nid, sid, val) name[cumulativePost[nid]+sid]=val;

#define SETPRE_INFO(name, nid, sid, val)  name[cumulativePre[nid]+sid]=val;

#define SETPOST_WEIGHTS(name, nid, sid, val)				\
	int _postInfo = postSynapticIds[cumulativePost[nid]+sid];		\
	int _nid_dest = (_postInfo&POST_SYN_NEURON_MASK);		\
	int _sid_dest = (_postInfo>>POST_SYN_NEURON_BITS)&(POST_SYN_NEURON_MASK);	\
	assert((_nid_dest>=0)&&(_nid_dest<N));					\
	assert((_sid_dest<(3*M))&&(_sid_dest>=0));					\
	int _pos = cumulativePre[_nid_dest]+ _sid_dest;			\
	assert((_pos>=0)&&(_pos < (3*N*(M+COMPACTION_ALIGNMENT_PRE))));					\
	name[_pos]=val;

	void CpuSNN::printMemoryInfo(FILE *fp)
	{
		fprintf(fp, "*************Memory Info *************\n");
		int totMemSize = networkInfoSize+synapticInfoSize+neuronInfoSize+spikingInfoSize;
		fprintf(fp, "Neuron Info Size:    %3.2f (%f MB)\n", neuronInfoSize*100.0/totMemSize,   neuronInfoSize/(1024.0*1024));
		fprintf(fp, "Synaptic Info Size:  %3.2f (%f MB)\n", synapticInfoSize*100.0/totMemSize, synapticInfoSize/(1024.0*1024));
		fprintf(fp, "Network Size:        %3.2f (%f MB)\n", networkInfoSize*100.0/totMemSize,  networkInfoSize/(1024.0*1024));
		fprintf(fp, "Firing Info Size:    %3.2f (%f MB)\n", spikingInfoSize*100.0/totMemSize,   spikingInfoSize/(1024.0*1024));
		fprintf(fp, "Additional Info:     %3.2f (%f MB)\n", addInfoSize*100.0/totMemSize,      addInfoSize/(1024.0*1024));
		fprintf(fp, "DebugInfo Info:      %3.2f (%f MB)\n", debugInfoSize*100.0/totMemSize,    debugInfoSize/(1024.0*1024));
		fprintf(fp, "**************************************\n\n");
	}

	void CpuSNN::resetWeights()
	{
		int j;

		//Reset weights to appropriate values...
		for(int g=0; g < numGrp; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				int offset = cumulativePre[i];
				for (j=0;j<Npre[i]; j++) wtChange[offset+j] = 0.0;									// synaptic derivatives
				for (j=0;j<Npre[i]; j++) synSpikeTime[offset+j] = SOME_LARGE_NEGATIVE_VALUE;	// some large negative value..

				//We have to set only the posts-weights...
				//synaptic weights are stored with respect to the post-synaptic neurons...
				for (j=0; j < Npost[i]; j++) {
					SETPOST_WEIGHTS(wt,i,j,grpInitWeights[g]);
				}
			}
		}

	}

	// initialize all the synaptic weights to appropriate values..
	// total size of the synaptic connection is 'length' ...
	void CpuSNN::initSynapticWeights()
	{
		preSynLength = cumulativePre[N-1]+Npre[N-1]+100;	// additional space for allowing some overflow...
		assert((preSynLength>=N*M)&&(preSynLength < (3*N*(M+COMPACTION_ALIGNMENT_PRE))));		// ensure preSynLength is not some garbage...

		// Initialize the network wtChange, wt, synaptic firing time
		wt  = new float[preSynLength];
		wtChange = new float[preSynLength];
		synSpikeTime = new int[preSynLength];
		synapticInfoSize = sizeof(float)*(preSynLength*3);

		resetWeights();

	}

	// We parallelly cleanup the postSynapticIds array to minimize any other wastage in that array by compacting the store
	// Appropriate alignment specified by ALIGN_COMPACTION macro is used to ensure some level of alignment (if necessary)
	void CpuSNN::compactConnections()
	{
		cumulativePost[0]=0;
		int lastCnt;
		for(int i=1; i < N; i++) {
			lastCnt = cumulativePost[i-1]+Npost[i-1]; //position of last pointer
			if (COMPACTION_ALIGNMENT_POST)
				lastCnt=lastCnt+COMPACTION_ALIGNMENT_POST-lastCnt%COMPACTION_ALIGNMENT_POST;
			cumulativePost[i]=lastCnt;
			assert(cumulativePost[i]<(N*(M+COMPACTION_ALIGNMENT_POST)));
		}

		// The number of synaptic connection at the input side might be random in nature
		// so the cumulativePre array stores the starting position of the pre-synaptic array
		// for each neuron....
		cumulativePre[0]=0;
		for(int i=1; i < N; i++) {
			lastCnt = cumulativePre[i-1]+Npre[i-1]; //position of last pointer
			if(COMPACTION_ALIGNMENT_PRE)
				lastCnt=lastCnt+COMPACTION_ALIGNMENT_PRE-lastCnt%COMPACTION_ALIGNMENT_PRE;
			cumulativePre[i]=lastCnt;
			assert(cumulativePost[i]<(3*N*(M+COMPACTION_ALIGNMENT_PRE)));
		}

		if (COMPACTION_ALIGNMENT_PRE)
			lastCnt=lastCnt+COMPACTION_ALIGNMENT_PRE-lastCnt%COMPACTION_ALIGNMENT_PRE;
		assert(lastCnt < (3*N*(M+COMPACTION_ALIGNMENT_PRE)));

		// compress the post_synaptic array according the new values of the cumulative counts....
		postSynCnt = cumulativePost[N-1]+Npost[N-1];
		int* tmp_post_syn = new int[postSynCnt+10];		// new buffer with required size + 10 bytes of additional space just to provide limited overflow
		networkInfoSize += sizeof(int)*(postSynCnt+10); // we keep track of the network info size
		for(int i=0; i < N; i++) {
			int lastCnt = cumulativePost[i];
			for( int j=0; j < Npost[i]; j++) {
				tmp_post_syn[lastCnt+j]=postSynapticIds[i*M+j];
				tmpSynapticDelay[lastCnt+j]=tmpSynapticDelay[i*M+j];
			}
		}

		delete[] postSynapticIds;							  // delete old buffer space

		// We no longer using the old statically allocated postSynapticIds connection array...
		networkInfoSize -= (sizeof(int)*N*M);

		postSynapticIds = tmp_post_syn;					  // copy the new required space

	}

	// Here we reorganize all the input connections to the neurons such that
	// the first set is all excitatory connection and next set is inhibitory connections
	void CpuSNN::regroupConnections(FILE *fp)
	{
		// Here we reorganize all the input connections to the neurons such that
		// the first set is all excitatory connection and next set is inhibitory connections
		for(int g=0; g < numGrp; g++) {
			for(int i=grpStartN[g]; i <= grpEndN[g]; i++) {
				assert(i<N);

				if((fp!=NULL)&&(i<100)) {
					fprintf(fp, " Before: %3d ( %3d ) : \t", i, Npre[i]);
					for(int j=0; j < Npre[i]; j++) {
						fprintf(fp, " %3d (%s) ", tmpPreSynapticIds[3*M*i+j]&POST_SYN_NEURON_MASK,
												 (tmpNeuronType[3*M*i+j]==INHIBITORY_NEURON)?"I":"E");
					}
					fprintf(fp, "\n");
				}

				int j=0;
				int k=Npre[i]-1;
				while((j < Npre[i])) {
					int inhibSID=-1,excitSID=-1;

					//TODO: comments ??
					while(j < Npre[i]) {
						/* found an inhibitory connection */
						if (tmpNeuronType[i*3*M+(j++)] == INHIBITORY_NEURON) {
							inhibSID = j-1;
							break;
						}
					}

					//TODO: comments ??
					while(k >= 0) {
						/* found an excitatory connection */
						if (tmpNeuronType[i*3*M+(k--)] == EXCITATORY_NEURON) {
							excitSID = k+1;
							break;
						}
					}

					// connections are already sorted out..
					if(j>k) break;

					//TODO: found a valid excitatory and inhibitory connection
					if ((inhibSID != -1)&&(excitSID != -1)&&(inhibSID<=excitSID)) {

						assert(inhibSID < Npre[i]);
						assert(excitSID < Npre[i]);

						tmpNeuronType[i*M*3+excitSID] = INHIBITORY_NEURON;
						tmpNeuronType[i*M*3+inhibSID] = EXCITATORY_NEURON;

						//inhibitory connection info..
						int tmp_inhi = tmpPreSynapticIds[i*M*3+inhibSID];
						int s_i = (tmp_inhi >> POST_SYN_NEURON_BITS);
						int n_i = (tmp_inhi &  POST_SYN_NEURON_MASK);

						//excitatory connection info...
						int tmp_exct = tmpPreSynapticIds[i*M*3+excitSID];
						int s_e = (tmp_exct >> POST_SYN_NEURON_BITS);
						int n_e = (tmp_exct &  POST_SYN_NEURON_MASK);

						//exchange the connection connection..
						//so that excitatory connection comes in the begining and inhibitory
						//connection in the end...
						tmpPreSynapticIds[i*M*3+inhibSID] = tmp_exct;
						tmpPreSynapticIds[i*M*3+excitSID] = tmp_inhi;

						//retrieve appropriate pre-synaptic connections..
						tmp_exct=postSynapticIds[cumulativePost[n_e]+s_e];
						tmp_inhi=postSynapticIds[cumulativePost[n_i]+s_i];

						if(((tmp_exct&POST_SYN_NEURON_MASK)!=i)||((tmp_inhi&POST_SYN_NEURON_MASK)!=i)) {
							fprintf(stderr,"i=%d, tmp_id1i=%d, tmp_id2i=%d\n", i, (tmp_exct&POST_SYN_NEURON_MASK),(tmp_inhi&POST_SYN_NEURON_MASK));
							fprintf(stderr,"s_i=%d, n_i=%d, s_e=%d, n_e=%d\n", s_i, n_i, s_e, n_e);
							fprintf(stderr,"inhibSID = %d, excitSID=%d, cum[n_i]=%d, cum[n_e]=%d\n", inhibSID, excitSID, cumulativePost[n_i], cumulativePost[n_e]);
							assert((tmp_exct&POST_SYN_NEURON_MASK)==i);
							assert((tmp_inhi&POST_SYN_NEURON_MASK)==i);
						}

						//exchange the post-synaptic connection because
						//we reorganizing the connections of excitatory and inhibitory synapses..
						postSynapticIds[cumulativePost[n_e]+s_e] = tmp_inhi;
						postSynapticIds[cumulativePost[n_i]+s_i] = tmp_exct;
					}
				}  // finished all synaptic connection in i

				if((fp!=NULL)&&(i<100)) {
					fprintf(fp, "After: %3d ( %3d ) : \t", i, Npre[i]);
					for(int j=0; j < Npre[i]; j++) {
						fprintf(fp, " %3d (%s) ", tmpPreSynapticIds[3*M*i+j]&POST_SYN_NEURON_MASK,
												 (tmpNeuronType[3*M*i+j]==INHIBITORY_NEURON)?"I":"E");
					}
					fprintf(fp, "\n");
				}
			} // finished all neurons in grpId..
		} // finished all groups...
	}

	void CpuSNN::updateParameters()
	{
		Ne = 0;
		Ni = 0;
		int state = 0;
		for(int g=0; g < numGrp; g++) {
			if (grpNeuronType[g]==INHIBITORY_NEURON) {
				Ni += grpSizeN[g];
				if(state == 0) {
					state = 1;
				}
			}
			else if(grpNeuronType[g]==EXCITATORY_NEURON) {
				Ne += grpSizeN[g];
				if(state != 0) {
					fprintf(stderr, "In this version of simulator...first create excitatory groups and then inhibitory groups\n");
					fprintf(stderr, "change the order of createGroup() so that first you create all excitatory groups\n");
					assert(0);
				}
			}
			else {
				fprintf(stderr, "Group %d has unknown property...\n", g);
				assert(grpNeuronType!=UNKNOWN_NEURON);
			}
		}
	}

	void CpuSNN::initFiringTables()
	{
		maxSpikesInhibitory = Ni*INHIBITORY_NEURON_MAX_FIRING_RATE;
		maxSpikesExcitatory = Ne*EXCITATORY_NEURON_MAX_FIRING_RATE;
		EfiringTable= new int[maxSpikesExcitatory];
		IfiringTable= new int[maxSpikesInhibitory];
		spikingInfoSize += sizeof(int)*((maxSpikesExcitatory+maxSpikesInhibitory) + 2*(1000+D+1));

	}

	// after all the initalization. Its time to create the synaptic weights, weight change and also
	// time of firing these are the mostly costly arrays so dense packing is essential to minimize wastage of space
	void CpuSNN::reorganizeNetwork()
	{
		//Double check...sometimes by mistake we might call reorganize network again...
		if(doneReorganization)
			return;

		fprintf(fpDotty, "\n\t}\n}\n");
		fclose(fpDotty);

		fprintf(stdout, "Beginning reorganization of network....\n");

		//update main set of parameters
		updateParameters();

		//assigning firing info
		initFiringTables();

		//..minimize any other wastage in that array by compacting the store
		compactConnections();

		// Here we reorganize all the input connections to the neurons such that
		// the first set is all excitatory connection and next set is inhibitory connections
		regroupConnections();

		// The post synaptic connections are sorted based on delay here
		reorganizeDelay();

		// Print statistics of the memory used to stdout...
		printMemoryInfo();

		// Print the statistics again but dump the results to a file
		printMemoryInfo(fpLog);

		// initialize the synaptic weights accordingly..
		initSynapticWeights();

		//not of much use currently
		updateRandomProperty();

		//printConnection("net.log");

		//printPreConnection();

		//ensure that we dont do all the above optimizations again
		doneReorganization = true;

		// if our initial operating mode is GPU_MODE, then it is time to
		// allocate necessary data within the GPU
		if(netInitMode==GPU_MODE)
			allocate();

		fprintf(stdout, "Starting GPU-SNN Simulations ....\n");
		delete[] tmpNeuronType;
		delete[] tmpPreSynapticIds;
		delete[] tmpSynapticDelay;

		//exit(0);
	}
