//
// Copyright (c) 2009 Regents of the University of California. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// 3. The names of its contributors may not be used to endorse or promote
//    products derived from this software without specific prior written
//    permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


	#include "config.h"
	#include "snn.h"
	
	// includes, library
	#include "cudpp/cudpp.h"

	// includes, project
	#include "cutil_inline.h"
	#include "cutil_math.h"
	
	#define TIMING_TABLE_COUNT   1000
	#define ROUNDED_TIMING_COUNT 1024  // (1000+D) rounded to multiple 128
	
	#define  SHARED_MEM_SIZE   (2048)
	#define  FIRE_CHUNK_CNT    (512)

    #define TESTING 0
    #define ENABLE_MORE_CHECK	0
	#define LOG_HALF_WARP_SIZE		(4)
	#define HALF_WARP_SIZE			(1 << LOG_HALF_WARP_SIZE)

	#define LOG_WARP_SIZE			(5)
	#define WARP_SIZE				(1 << LOG_WARP_SIZE)

	#define MAX_NUM_THREADS			512

    ///////////////////////////////////////////////////////////////////
    // Some important ideas that explains the GPU execution are as follows:
    //  1. Each GPU block has a local firing table (called fireTable). The block of threads
    //     reads a bunch of neurons parameters and determines if it needs to fire or not
    //     Whenever a neuron need to fire, it keeps track of the fired neuron in the local
    //     table. When the table is full, we go and write back the fireTable to the global
    //     firing table. 
    //  2. Firing information is maintained in two tables globally (timingTable and the globalFiringTable)
    //     for excitatory neuron population and inhibitory neurons.
    //     The globalFiringTable only stores a sequence of id corresponding to fired neurons.
    //     The timingTable store the total number of fired neurons till the current time step t.
    //     These two tables are flushed and adjusted every second.
    //     This approach requires about half of the memory compared to the traditional AER scheme which
    //     stores the firing time and firing id together.
    //  For more details kindly read the enclosed report (report.pdf) in the source directory
    ///////////////////////////////////////////////////////////////////

	#define GPU_LTP(t)   (gpuInfo.ALPHA_LTP*__expf(-(t)/gpuInfo.TAU_LTP))
	#define GPU_LTD(t)   (gpuInfo.ALPHA_LTD*__expf(-(t)/gpuInfo.TAU_LTD))
	
	__device__ int  timingTableE[ROUNDED_TIMING_COUNT];
	__device__ int  timingTableI[ROUNDED_TIMING_COUNT];
	__device__ int	secEfireCnt=0;
	__device__ int	secIfireCnt=0;
	__device__ int  prevSecEfireCnt=0;
	__device__ int  prevSecIfireCnt=0;
	__device__ int  totEfireCnt=0;
	__device__ int  totIfireCnt=0;

	__device__ int  generatedSpikesE=0;
	__device__ int  generatedSpikesI=0;
	__device__ int  receivedSpikesE=0;
	__device__ int  receivedSpikesI=0;
	__device__ int  senderIdE[1000];
	__device__ int  senderIdI[1000];
	__device__ int  receiverId[1000];

	__device__ __constant__ network_ptr_t		gpuPtrs;
	__device__ __constant__ network_info_t		gpuInfo;

	__device__ __constant__ noiseGenProperty_t*	gpu_noiseGenGroup;

	__device__ __constant__ float				constData[256];

	float data[256];	

    texture <int, 1, cudaReadModeElementType> timingTableE_tex;
	texture <int, 1, cudaReadModeElementType> timingTableI_tex;
	
	#define MAX_NUM_BLOCKS 200
	#define LOOP_CNT		10
	
	#define NO_KERNEL_ERRORS				 0xc00d
	
	#define NEW_FIRE_UPDATE_OVERFLOW_ERROR1  0x61
	#define NEW_FIRE_UPDATE_OVERFLOW_ERROR2  0x62

	
	__device__ int	 tmp_val[MAX_NUM_BLOCKS][LOOP_CNT];
	__device__ int	 retErrCode=NO_KERNEL_ERRORS;
	__device__ float retErrVal[MAX_NUM_BLOCKS][10];

	#define INIT_CHECK(enable, src, val)			\
		{											\
			if(enable)								\
			  if(threadIdx.x==0) 					\
				src = val;							\
		}

	#define ERROR_CHECK(enable, src, val)			\
		{											\
			if(enable) 								\
				if( blockIdx.x >= MAX_NUM_BLOCKS) { \
					src = val;						\
					return;							\
				}									\
		}		


	#define UPDATE_ERROR_CHECK4(src, val1, val2, val3, val4)			\
		{																\
				if((threadIdx.x==0)&&(blkErrCode))	{					\
				retErrCode = src;										\
				if(ENABLE_MORE_CHECK) {									\
				retErrVal[blockIdx.x][0]=0xdead;						\
				retErrVal[blockIdx.x][1]=4;		 						\
				retErrVal[blockIdx.x][2]=val1;							\
				retErrVal[blockIdx.x][3]=val2;							\
				retErrVal[blockIdx.x][4]=val3;							\
				retErrVal[blockIdx.x][5]=val4;							\
				}														\
				return;													\
			}															\
		}
		
	#define	MEASURE_LOADING			0

	#define MEASURE_GPU(pos,val)					\
		{											\
			if(MEASURE_LOADING)						\
				if (threadIdx.x==0)					\
					tmp_val[blockIdx.x][pos] += val;\
		}

	// example of the quick synaptic table
	// index     cnt
	// 0000000 - 0
	// 0000001 - 0
	// 0000010 - 1
	// 0100000 - 5
	// 0110000 - 4
	int tableQuickSynId[256];
	__device__ int  gpu_tableQuickSynId[256];
	void initTableQuickSynId()
	{
	   void* devPtr;
	   
	   for(int i=1; i < 256; i++) {
		 int cnt=0;
		 while(i) {
		   if(((i>>cnt)&1)==1) break;
		   cnt++;
		   assert(cnt<=7);
		 }
		 tableQuickSynId[i]=cnt;		 
	   }
	   
	   cudaGetSymbolAddress(&devPtr, gpu_tableQuickSynId);
	   CUDA_SAFE_CALL( cudaMemcpy( devPtr, tableQuickSynId, sizeof(tableQuickSynId), cudaMemcpyHostToDevice));            

	}
	
    // KERNEL: 
	//  In this kernel we return some important parameters, data values of 
	//  the initialized SNN network using the retErrVal array. 
	//  This is to just to ensure that the initialization has been
	//  done correctly and we can concentrate on the actual kernel code
	//  for bugs rather than errors due to incorrect initialization.
	__global__ void check_GPU_init ()
	{
		int i=1;

		if(threadIdx.x==0 && blockIdx.x==0) {	
			float hstep = gpuInfo.maxWeight/(HISTOGRAM_SIZE);
			retErrVal[0][i++]= gpuInfo.N;
			retErrVal[0][i++]= gpuInfo.M;
			retErrVal[0][i++]= gpuInfo.D;
			retErrVal[0][i++]= gpuInfo.Ne;
			retErrVal[0][i++]= gpuInfo.Ni;			
			retErrVal[0][i++]= gpuPtrs.wt[0];
           int histPos_wt = (int) ((gpuPtrs.wt[0]/hstep));
		   int test = 0;
				  	if (histPos_wt  >= HISTOGRAM_SIZE)
						test = 1; 
					else if(histPos_wt  < 0)
						test = 2;
					else				
						test = 3;
			retErrVal[0][i++]= histPos_wt;			
			retErrVal[0][i++]= test;
			retErrVal[0][i++]= gpuInfo.numRandNeurons;
			retErrVal[0][0]  = i;
			i = 1;
			retErrVal[1][i++]= tex1Dfetch(timingTableE_tex,0);
			retErrVal[1][i++]= tex1Dfetch(timingTableE_tex,1);
			retErrVal[1][0]  = i;
		}
	}
	// KERNEL: After every iteration we update the timing table
	// so that we have the new values of the fired neurons for the
	// current time t.
	__global__ void gpu_timingTableUpdate(int t)
	{
	   if ( threadIdx.x == 0 && blockIdx.x == 0) {
			timingTableE[t+gpuInfo.D+1]  = secEfireCnt;
			timingTableI[t+gpuInfo.D+1]  = secIfireCnt;
	   }
	   __syncthreads();									     
	}

	// KERNEL: This kernel is called after initialization of various parameters
	// so that we can reset all required parameters. 
	__global__ void kernel_init ()
	{
		if(threadIdx.x==0 && blockIdx.x==0) {
			for(int i=0; i < ROUNDED_TIMING_COUNT; i++) {
				timingTableE[i]   = 0;   
				timingTableI[i]   = 0;   
			}
		}
	}

	// KERNEL: This is useful for generating suitable thalamic inputs
	// to the network. The ids of neurons to be used is stored in the randId array and 
	// the kernel reads the randId array to initialize each input thalamic current
	// of the selected neurons to appropriate values..
	// future modification is to generate the random numbers using GPU itself instead of
	// creating the random number at the CPU side and doing a memcopy operation.
	__global__ void kernel_initThalInput()
	{
	   const int tid = threadIdx.x;
	   const int bid = blockIdx.x;
	   
	   const int idBegin = bid*blockDim.x + tid;
	   const int idStep  = blockDim.x*gridDim.x;	   
	   int accPos = 0;
	   noiseGenProperty_t*	gpu_noiseGenGroup = (noiseGenProperty_t *) gpuPtrs.noiseGenProp;
	   for (int i=0; i < gpuInfo.numNoise; i++)
	   {
		 int    noiseGrpId       = gpu_noiseGenGroup[i].groupId;
		 float  currentStrength  = gpu_noiseGenGroup[i].currentStrength;
		 float  neuronPercentage = gpu_noiseGenGroup[i].neuronPercentage;
		 int    ncount = gpu_noiseGenGroup[i].ncount;
		 int    nstart = gpu_noiseGenGroup[i].nstart;			 
		 int    nrands = gpu_noiseGenGroup[i].rand_ncount;

		 int randCnt = accPos;
		 for(int j=idBegin; j < nrands; j+=idStep) {
			 int randId = gpuPtrs.randId[randCnt+j];
			 // fprintf(fpLog, " %d %d \n", simTimeSec*1000+simTimeMs, randNeuronId[randCnt]);
			 // assuming there is only one driver at a time.
			 // More than one noise input is not correct... 			 
			 gpuPtrs.I[randId] = currentStrength;
		 }
		 accPos += nrands;
	   }
	}

	///////////////////////////////////////////////////////////////////////////
	/// Device local function:      gpu_newFireUpdate
	/// Description: 1. Copy neuron id from local table to global firing table.
	///				 2. Reset all neuron properties of neuron id in local table
	//
	/// fireTablePtr:  local shared memory firing table with neuron ids of fired neuron
	/// EfireCnt:      number of excitatory neurons in local table that has fired
	/// IfireCnt:      number of inhibitory neurons in lccal table that has fired
	/// simTime:      current global time step..stored as neuron firing time  entry
	/////////////////////////////////////////////////////////////////////////

	__device__ void gpu_newFireUpdate(	int* fireTablePtr,
										int EfireCnt,
										int IfireCnt,
										int simTime)
	{		
		__shared__ volatile int cntE;	
		__shared__ volatile int cntI;
		__shared__ volatile int blkErrCode;

		INIT_CHECK(true, blkErrCode, 0);
		    
		if (0==threadIdx.x) {
			
			// get a distinct counter to store firing info into the firing table
			cntE = atomicAdd(&secEfireCnt, EfireCnt);
			cntI = atomicAdd(&secIfireCnt, IfireCnt);

			MEASURE_GPU(1, EfireCnt);
			MEASURE_GPU(2, IfireCnt);
			
			//check for overflow in the firing table size....
			if((cntE+EfireCnt)>gpuInfo.maxSpikesExcitatory) {
				blkErrCode = NEW_FIRE_UPDATE_OVERFLOW_ERROR1;
			}
			else if((cntI+IfireCnt)>gpuInfo.maxSpikesInhibitory) {
				blkErrCode = NEW_FIRE_UPDATE_OVERFLOW_ERROR2;		
			}
		}

		__syncthreads();

		UPDATE_ERROR_CHECK4(blkErrCode,cntE,EfireCnt,cntI,IfireCnt);
		
		for (int i=0; i < (EfireCnt+IfireCnt); i+=(blockDim.x)) {
			// check to ensure my index is within the range of valid address..
			if ((i+threadIdx.x) < (EfireCnt+IfireCnt)) {
				// read the firing id from the local table..
				int nid  = fireTablePtr[i+threadIdx.x];

				// get suitable position from the 
				int pos;
				if(nid < gpuInfo.Ne) {
				  pos = atomicAdd((int*)&cntE, 1);
  				  gpuPtrs.EfiringTable[pos]  = nid;
				}
				else {
				  pos = atomicAdd((int*)&cntI, 1);
 				  gpuPtrs.IfiringTable[pos]  = nid;
				}
				// TODO: convert this to use coalesced access by grouping into a 
				// single 16 byte access. This might improve bandwidth performance
				// This is fully uncoalsced access...need to convert to coalsced access..
				gpuPtrs.v[nid] = -65.0f;
				gpuPtrs.u[nid] += gpuPtrs.d[nid];
				gpuPtrs.lastSpikeTime[nid] = simTime;
			 }
		}
		__syncthreads();
	}

	// change this with selective upgrading
	// technique used for firing neurons
	__global__ void gpu_globalStateUpdate()
	{
		int i;
		__shared__ volatile int sharedTemp;
		__shared__ int sh_tableQuickSynId[256];
		__shared__ int sh_tmp_I_set;
		
		//__shared__ volatile float sh_I_sum[MAX_NUM_THREADS];	
		float I_sum;
		
		for(int i=0; i < 256; i+=blockDim.x){
		   if((i+threadIdx.x) < 256){
			  sh_tableQuickSynId[i+threadIdx.x]=gpu_tableQuickSynId[i+threadIdx.x];
		   }
		}
		
#define LOG_CURRENT_GROUP 5
#define CURRENT_GROUP	  ( 1 << LOG_CURRENT_GROUP )

		int gid=threadIdx.x + blockDim.x*blockIdx.x; //global id
		for(i=0; i < gpuInfo.N; i+= (blockDim.x*gridDim.x)) {
			ERROR_CHECK(ENABLE_MORE_CHECK, retErrCode, MAX_NUM_BLOCKS);
			int   pos = i + gid;
			if (pos<gpuInfo.N) {
			   I_sum =  gpuPtrs.I[pos];
	#if 1			   
			   int lmt = gpuPtrs.Npre[pos];
			   int cum_pos = gpuPtrs.cumulativePre[pos];
			   for(int j=0; j <= (lmt>>LOG_CURRENT_GROUP); j++) {
			      // because of malloc2D operation we are using pitch, pos, j to get
				  // actual position of the input current..
			      int* tmp_I_set_p = ((int*)((char*)gpuPtrs.I_set + j * gpuInfo.I_setPitch) + pos);
				  int tmp_I_set    = *tmp_I_set_p;
	#if 1
				  //table lookup based find bits that are set
				  int cnt=0;
				  int tmp_I_cnt = 0;
				  while(tmp_I_set) {					 
					 int k=(tmp_I_set>>(8*cnt))&0xff;
					 if(k==0) { cnt = cnt+1; continue; }
					 int wt_i = sh_tableQuickSynId[k];				 
					 float wt = gpuPtrs.wt[cum_pos + j*32+ cnt*8 + wt_i]; 
					 int id;
					 if(TESTING) {
						 if(wt < 0.0)  id = atomicAdd(&receivedSpikesI, 1);
						 else    	  id = atomicAdd(&receivedSpikesE, 1);
					 }
					 I_sum +=  wt;					 
					 tmp_I_cnt++;
					 tmp_I_set = tmp_I_set & (~(1<<(8*cnt+wt_i)));
				  }
				  // reset the input if there are any bit'wt set
				  if(tmp_I_cnt) *tmp_I_set_p = 0;
	#endif		
				  __syncthreads();
				  				  
			   }
	#endif	   
				__syncthreads();
				gpuPtrs.I[pos] = 0;
				gpuPtrs.v[pos]+=0.5f*((0.04f*gpuPtrs.v[pos]+5)*gpuPtrs.v[pos]+140-gpuPtrs.u[pos]+I_sum); //sh_I_sum[threadIdx.x]); // for numerical stability			
				gpuPtrs.v[pos]+=0.5f*((0.04f*gpuPtrs.v[pos]+5)*gpuPtrs.v[pos]+140-gpuPtrs.u[pos]+I_sum); //sh_I_sum[threadIdx.x]); // time step is 0.5ms
				gpuPtrs.u[pos]+=gpuPtrs.a[pos]*(0.2f*gpuPtrs.v[pos]-gpuPtrs.u[pos]);
		   }
		}
	}


	/************************ VARIOUS KERNELS IN THE SYSTEM *********************************/

	///////////////////////////////////////////////////////////////////////////
	/// Device local function:      gpu_updateLTP
	/// Description:                Computes the STDP update values for each of fired
	///                             neurons stored in the local firing table.
	///
	/// fireTablePtr:  local shared memory firing table with neuron ids of fired neuron
	/// fireCnt:      number of fired neurons in local firing table
	/// simTime:     current global time step..stored as neuron firing time  entry
	///////////////////////////////////////////////////////////////////////////

	// synaptic grouping for LTP Calculation
	#define		LTP_GROUPING_SZ     16

	__device__ void gpu_updateLTP(int* fireTablePtr, 
								  int fireCnt,
								  int simTime)
	{
		for(int i=0; i < fireCnt; i+= (blockDim.x/LTP_GROUPING_SZ))
		{
		  int pos = i+(threadIdx.x/LTP_GROUPING_SZ);
		  if( pos < fireCnt ) {
			  int	 nid	 = fireTablePtr[pos];
			  int    tmp_pre = gpuPtrs.Npre_exc[nid];
			  int    tmp_cum = gpuPtrs.cumulativePre[nid];
			  for(int j=tmp_cum; j < (tmp_cum+tmp_pre); j+=LTP_GROUPING_SZ) {
				int pos = j + (threadIdx.x%LTP_GROUPING_SZ);
				if(pos < (tmp_cum+tmp_pre))
					gpuPtrs.wtChange[pos] += GPU_LTP(simTime-gpuPtrs.synSpikeTime[pos]);
			  }
		  }
		}
		__syncthreads();
	}

	// KERNEL DESCRIPTION:
	//   This is the main kernel that finds the list of fired neuron.
	//   Each thread reads a unique neuron value and checks if it has fired or not.
	//   If the neuron has fired, we keep the id in a local table "fireTable".
	//   If the "fireTable" need to be flushed, we call the functions, "gpu_newFireUpdate" and "gpu_updateLTP"
	//   "gpu_newFireUpdate" writes the fired id to the global table and update the states of fired neurons.
	//   "gpu_updateLTP" updates synaptic weights according to the STDP rule.
	__global__
	void kernel_findFiring (int t, int sec, int simTime, int gpu_with_stdp)
	{
		__shared__ volatile int fireCnt;
		__shared__ volatile int IfireCnt;
		__shared__ int fireTable[FIRE_CHUNK_CNT];

		if (0==threadIdx.x) {
			fireCnt	 = 0; // initialize total cnt to 0
			IfireCnt = 0; // initialize inh. cnt to 0
		}
		
		// ignore this unless you are doing real debugging gpu code...
		INIT_CHECK(ENABLE_MORE_CHECK, retErrCode, NO_KERNEL_ERRORS);	
		
		const int gid=blockIdx.x*blockDim.x+threadIdx.x;
		const int gnthreads=blockDim.x*gridDim.x;

		__syncthreads();
		
		for (int i=blockDim.x*blockIdx.x; i < gpuInfo.N; i += gnthreads) {

			int  myI = i + threadIdx.x;
			bool needToWrite = false;

			// neuron fired
			if (myI < gpuInfo.N)
				needToWrite = (gpuPtrs.v[myI] >= 30.0);													

			int fireId;

			// if fired obtain the required id
			if(needToWrite)
				fireId = atomicAdd((int*)&fireCnt, 1);

			if (needToWrite && (fireId<(FIRE_CHUNK_CNT))) {
				// store ID of the fired neuron
				fireTable[fireId] = myI; 	
					
				// fired neuron is inhibitory
				if((myI >= gpuInfo.Ne)) {
					atomicAdd((int*)&IfireCnt, 1); 	
				}
				needToWrite = 0;
			}
			
			__syncthreads();
			
			// table is full.. dump the local table to the global table before proceeding
			if (fireCnt >= (FIRE_CHUNK_CNT)) {
			
				// clear the table and update...
				gpu_newFireUpdate(fireTable, (FIRE_CHUNK_CNT-IfireCnt), IfireCnt, simTime);
			
				// update based on stdp rule
				if(gpu_with_stdp && gpuInfo.STDPenabled)
					gpu_updateLTP(fireTable, (FIRE_CHUNK_CNT), simTime);

				// reset counters
				if (0==threadIdx.x) {
					IfireCnt = 0;
					fireCnt  = 0;
					MEASURE_GPU(MEASURE_LOADING, 1);
				}

				__syncthreads();

				if(needToWrite) {
					fireId = atomicAdd((int*)&fireCnt, 1);
					if((myI >= gpuInfo.Ne))
						atomicAdd((int*)&IfireCnt, 1); 
					fireTable[fireId] = myI;
				}
			}
		}

		if(fireCnt) {
		   gpu_newFireUpdate(fireTable, (fireCnt-IfireCnt), IfireCnt, simTime);
		   if(gpu_with_stdp && gpuInfo.STDPenabled)
			gpu_updateLTP(fireTable, (fireCnt), simTime);
		}
	}


	// KERNEL DETAILS:
	//   This kernel is called every second to adjust the timingTable and globalFiringTable
	//   We do the following thing:
	//   1. We discard all firing information that happened more than 1000-D time step.
	//   2. We move the firing information that happened in the last 1000-D time step to 
	//      the begining of the gloalFiringTable. 
	//   3. We read each value of "wtChange" and update the value of "synaptic weights wt".
	//      We also clip the "synaptic weight wt" to lie within the required range.
	#define UPWTS_GROUPING_SZ		32
	#define UPDATE_WEIGHTS_ERROR1	0x80
	__global__ void kernel_updateWeightsFiring()
	{
		int gid=blockIdx.x*blockDim.x+threadIdx.x;
		int gnthreads=blockDim.x*gridDim.x;
		__shared__ volatile int errCode;

		// CHECK !!!!
		// shift the firing table so that the initial information in
		// the firing table contain the firing information for the last D time step
		for(int p=timingTableE[980+gpuInfo.D+1],k=0;p<timingTableE[999+gpuInfo.D+1];p+=(2*gnthreads),k+=(2*gnthreads)) {

			if((p+gid)<timingTableE[999+gpuInfo.D+1])
				gpuPtrs.EfiringTable[k+gid]=gpuPtrs.EfiringTable[p+gid];

			if((p+gid+gnthreads)<timingTableE[999+gpuInfo.D+1])
				gpuPtrs.EfiringTable[k+gid+gnthreads]=gpuPtrs.EfiringTable[p+gid+gnthreads];

		}

		const int threadIdGrp       = (threadIdx.x%UPWTS_GROUPING_SZ);
		const int grpId				= (gid/UPWTS_GROUPING_SZ);
		const int grpNCnt			= (gnthreads/UPWTS_GROUPING_SZ) + ((gnthreads%UPWTS_GROUPING_SZ)!=0);	

		// update the synaptic weights from the synaptic weight derivatives
		for(int i=grpId; i < gpuInfo.N; i+=grpNCnt) {
			int eCnt = gpuPtrs.Npre_exc[i];
			int nCum = gpuPtrs.cumulativePre[i];
			// synaptic grouping
			for(int j=nCum; j < (nCum+eCnt); j+=UPWTS_GROUPING_SZ) {
				//excitatory connection change the synaptic weights
				int jpos=j+threadIdGrp;
				if(jpos < (nCum+eCnt)) {
					float t_wt   = gpuPtrs.wt[jpos];
					float t_wtChange  = gpuPtrs.wtChange[jpos];
					t_wt+=0.01f+t_wtChange;
					t_wtChange*=0.90f;
					if(t_wt>gpuInfo.maxWeight) t_wt=gpuInfo.maxWeight;
					if(t_wt<0)  t_wt=0.0f;
					gpuPtrs.wt[jpos]=t_wt;
					gpuPtrs.wtChange[jpos]=t_wtChange;
				}
			}
		}
	}

	// KERNEL DESCRIPTION:
	// This is the second part of the previous kernel "kernel_updateWeightsFiring"
	// After all the threads/blocks had adjusted the firing table and the synaptic weights,
	// we update the timingTable so that the firing information that happended in the last D
	// time step would become the first D time step firing information for the next cycle of simulation.
	// We also reset/update various counters to appropriate values as indicated in the second part 
	// of this kernel.
	__global__ void kernel_updateWeightsFiring2()
	{
		// CHECK !!!
		int D = gpuInfo.D;
		// reset the firing table so that we have the firing information
		// for the last D time step  to be used in the next cycle of the simulation
		if(blockIdx.x==0) {
			for(int i=threadIdx.x; i < D; i+=blockDim.x) {
				timingTableE[i+1] = timingTableE[980+D+1+i]-timingTableE[980+D+1];
				timingTableI[i+1] = timingTableI[980+D+1+i]-timingTableI[980+D+1];
			}
		}

		__syncthreads();

		// reset various counters for the firing information
		if((blockIdx.x==0)&&(threadIdx.x==0)) {
			timingTableI[gpuInfo.D]  = 0;
			totEfireCnt		+= secEfireCnt;
			totIfireCnt		+= secIfireCnt;
			prevSecEfireCnt = secEfireCnt;
			prevSecIfireCnt = secIfireCnt;
			secEfireCnt		= timingTableE[gpuInfo.D];
			secIfireCnt		= 0;
		}
	}

	// kernel to generate the synatpic weight histogram...
	__device__ int gpu_sHist[HISTOGRAM_SIZE+2];
	__device__ int gpu_sdHist[HISTOGRAM_SIZE+2];
	__global__ void kernel_getHistogram()
	{
		__shared__ int sHist[HISTOGRAM_SIZE+2];
		__shared__ int sdHist[HISTOGRAM_SIZE+2];	

		for(int i=threadIdx.x; i < (HISTOGRAM_SIZE+2); i+=blockDim.x) {
				sHist[i]=0;
				sdHist[i]=0;
		}

		__syncthreads();

		int gid=blockIdx.x*blockDim.x+threadIdx.x;
		int gnthreads=blockDim.x*gridDim.x;		
		float sStep  = gpuInfo.maxWeight/(HISTOGRAM_SIZE);
		float sdStep = gpuInfo.maxWeightChange/(HISTOGRAM_SIZE);

		int histSize = HISTOGRAM_SIZE;

		for(int i=blockIdx.x; i < gpuInfo.N; i+=gridDim.x) {
			int pos = gpuPtrs.cumulativePre[i];
			for(int j=threadIdx.x; j < gpuPtrs.Npre_exc[i]; j+=blockDim.x) {
				  int histPos_wt = (int) ((gpuPtrs.wt[pos+j]/sStep));
				  int histPos_wtChange = (int) (((gpuPtrs.wtChange[pos+j]+(gpuInfo.maxWeightChange/2))/sdStep));

				  	if (histPos_wt  >= histSize)
						atomicAdd(&sHist[histSize+1],1);//overflows...
					else if(histPos_wt  < 0)
						atomicAdd(&sHist[0],1);			//underflows...
					else				
						atomicAdd(&sHist[histPos_wt+1],1);//within range

					if(histPos_wtChange >= histSize) 
						atomicAdd(&sdHist[histSize+1],1);//overflows...						
					else if(histPos_wtChange < 0)						
						atomicAdd(&sdHist[0],1);		 //underflows...
					else
						atomicAdd(&sdHist[histPos_wtChange+1],1);//within range				 
			}
		}

		__syncthreads();

		if(threadIdx.x==0) {
			for(int i=0; i < histSize+2; i++) {
				atomicAdd(&gpu_sHist[i],  sHist[i]);
				atomicAdd(&gpu_sdHist[i], sdHist[i]);
			}
		}
		__syncthreads();
	}

    #define NUM_THREADS 128
	#define EXCIT_READ_CHUNK_SZ		(NUM_THREADS>>1)
	#define EXCIT_GROUPING_SZ		(WARP_SIZE)

	#define CURRENT_UPDATE_ERROR1   0x41
	#define CURRENT_UPDATE_ERROR2   0x42
	#define CURRENT_UPDATE_ERROR3   0x439
	#define CURRENT_UPDATE_ERROR4   0x44
	#define CURRENT_UPDATE_ERROR5   0x45

    // KERNEL DESCRIPTION:
	//  This kernel is required for updating and generating spikes on all 
	//  excitatory connection from the fired neuron. The LTD computation
	//  is also executed by this approach kernel.
	__global__ void gpu_doCurrentUpdateExcit(int simTimeMs, int simTimeSec, int simTime, int gpu_with_stdp)
	{
		__shared__			int sh_neuronOffsetTable[EXCIT_READ_CHUNK_SZ+2];
		__shared__			int sh_delayLength[EXCIT_READ_CHUNK_SZ+2];
		__shared__			int sh_delayIndexStart[EXCIT_READ_CHUNK_SZ+2];
		__shared__ volatile int sh_excNeuronCnt;

		const int threadIdGrp	= (threadIdx.x%EXCIT_GROUPING_SZ);
		const int grpId			= (threadIdx.x/EXCIT_GROUPING_SZ);	
		int updateCnt	  = 0;

		__shared__ volatile int sh_blkErrCode;

		if(ENABLE_MORE_CHECK) {			
			if(threadIdx.x<=0) 
				sh_blkErrCode = 0;		
		}

		// this variable is used to note down
	    // number of updates done by different blocks
		if(threadIdx.x<=0) 
			sh_excNeuronCnt = 0;

      	__syncthreads();
		
		// stores the number of fired neurons at time t
		int k      = secEfireCnt - 1;

		// stores the number of fired neurons at time (t - D)
		int k_end  = tex1Dfetch(timingTableE_tex,simTimeMs+1);

		int t_pos  = simTimeMs;

		// we need to read (k-k_end) neurons from the firing 
		// table and do necesary updates for all these post-synaptic
		// connection in these neurons..
		while((k>=k_end) &&(k>=0)) {

			// at any point of time EXCIT_READ_CHUNK_SZ neurons
			// read different firing id from the firing table
			if (threadIdx.x<EXCIT_READ_CHUNK_SZ) {
				int fPos = k - (EXCIT_READ_CHUNK_SZ*blockIdx.x) - threadIdx.x; 
				if ((fPos >= 0) && (fPos >= k_end)) {
   			        int id = gpuPtrs.EfiringTable[fPos];

					// find the time of firing based on the firing number fPos
					while ( !((fPos >= tex1Dfetch(timingTableE_tex, t_pos+gpuInfo.D)) 
						   && (fPos <  tex1Dfetch(timingTableE_tex, t_pos+gpuInfo.D+1)))) {
                        t_pos = t_pos - 1;
						if(ENABLE_MORE_CHECK) {
							if (!((t_pos+gpuInfo.D) >= 0)) {
								int i=2;
								sh_blkErrCode = CURRENT_UPDATE_ERROR3;								
								retErrVal[blockIdx.x][0] = 0xdead;
								retErrVal[blockIdx.x][i++] = simTimeMs;
								retErrVal[blockIdx.x][i++] = t_pos;
								retErrVal[blockIdx.x][i++] = fPos;
								retErrVal[blockIdx.x][i++] = tex1Dfetch(timingTableE_tex, t_pos+gpuInfo.D-1);
								retErrVal[blockIdx.x][i++] = tex1Dfetch(timingTableE_tex, t_pos+gpuInfo.D);
								retErrVal[blockIdx.x][1]   = i;
								break;
							}
						}
					}
					
					// find the time difference between firing of the neuron and the current time
					int tD  = simTimeMs - t_pos;

					// find the various delay parameters for neuron 'id', with a delay of 'tD'
					int tPos = (gpuInfo.D+1)*id+tD;					
					sh_neuronOffsetTable[threadIdx.x]= gpuPtrs.cumulativePost[id];
					sh_delayLength[threadIdx.x]      = gpuPtrs.postDelayInfo[tPos].delay_length;
					sh_delayIndexStart[threadIdx.x]  = gpuPtrs.postDelayInfo[tPos].delay_index_start;

					// this is to indicate that the current thread 
					// has a valid delay parameter for post-synaptic firing generation
					atomicAdd((int*)&sh_excNeuronCnt,1);
				}
			}
			
			__syncthreads();

			if(ENABLE_MORE_CHECK)
				if (sh_blkErrCode) break;

			// if cnt is zero than no more neurons need to generate
			// post-synaptic firing, then we break the loop.
			int cnt = sh_excNeuronCnt;
			updateCnt += cnt;
			if(cnt==0)  break;

			// first EXCIT_GROUPING_SZ threads e the post synaptic
			// firing for first neuron, and so on. each of this group
			// needs to generate (M/D) spikes for every fired neuron, every second
			// for M=500,D=20, we need to generate 25 spikes for each fired neuron
			// for M=600,D=20, we need to generate 30 spikes for each fired neuron 
			for(int pos=grpId; pos < cnt; pos += (NUM_THREADS/EXCIT_GROUPING_SZ)) {

				int offset = sh_neuronOffsetTable[pos];

				int d   = sh_delayIndexStart[pos] + threadIdGrp;

				if (!(threadIdGrp < sh_delayLength[pos])) 
					continue;

				int post_info = gpuPtrs.postSynapticIds[offset+d];

				// get neuron id
				int p_i = (post_info&POST_SYN_NEURON_MASK);

				// get synaptic id
				int wt_i = (post_info>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK;
				
				// get the actual position of the synapses and other variables...
				int pos_ns = gpuPtrs.cumulativePre[p_i] + wt_i;
				
				if(ENABLE_MORE_CHECK) {
					if(p_i >= gpuInfo.N) sh_blkErrCode = CURRENT_UPDATE_ERROR1;
					if(wt_i >= gpuPtrs.Npre[p_i]) sh_blkErrCode =CURRENT_UPDATE_ERROR2;				
					if (sh_blkErrCode) break;
				}
				
				// approach taken from CUDA manual for doing 2D memory access using pitch and position
			    int* tmp_I_set_p = ((int*)((char*)gpuPtrs.I_set + (wt_i>>5) * gpuInfo.I_setPitch) + p_i);
				atomicOr(tmp_I_set_p, 1 << (wt_i%32));
				//atomicOr(&gpuPtrs.I_set[wt_i/32][p_i], (1<<(wt_i%32))); //uncoalesced access
				//gpu_I_set[wt_i/32][p_i] += (1<<(wt_i%32));

				gpuPtrs.synSpikeTime[pos_ns] = simTime;		  //uncoalesced access

				if(TESTING) int id = atomicAdd(&generatedSpikesE, 1);
				//senderIdE[id] = post_info;

				if(gpuInfo.STDPenabled && gpu_with_stdp)  {
					int stdp_tDiff = simTime-gpuPtrs.lastSpikeTime[p_i];
					if((stdp_tDiff/gpuInfo.TAU_LTD)<25)
					   gpuPtrs.wtChange[pos_ns] -= GPU_LTD(stdp_tDiff);	  //uncoalesced access
				}
				
			} //(for all excitory neurons in table)
		
			__syncthreads();

			if(threadIdx.x==0)  
				sh_excNeuronCnt = 0;

	        if(ENABLE_MORE_CHECK) 
			   if(sh_blkErrCode) break;

			k = k - (gridDim.x*EXCIT_READ_CHUNK_SZ);

			__syncthreads();

		}

		__syncthreads();

		MEASURE_GPU(3, updateCnt);

	    if(ENABLE_MORE_CHECK)	
		   if (sh_blkErrCode) {
			     retErrCode = sh_blkErrCode;
			     return;
		   }	
		// end of excitory part....Do we want a separate kernel for 
		// doing the inhibitory part ???
	}

	#define ICURRENT_UPDATE_ERROR1   0x51
	#define ICURRENT_UPDATE_ERROR2   0x52
	#define ICURRENT_UPDATE_ERROR3   0x53
	#define ICURRENT_UPDATE_ERROR4   0x54
	#define ICURRENT_UPDATE_ERROR5   0x55

	#define INHIB_GROUPING_SZ		(WARP_SIZE)

	// KERNEL DESCRIPTION:
	//  This kernel is required for updating and generating spikes on all 
	//  inhibitory connection from the fired neuron. This function looks
	//  mostly like the previous kernel but has been optimized for inhibitory neurons
	//  because they have a fixed delay of 1ms. Ultimately we will merge this
	//  kernel with the previous kernel.
	__global__ void gpu_doCurrentUpdateInhib(int cur_time, int sec, int simTime)
	{	
		__shared__ volatile	int sh_inhNeuronCnt;
		__shared__ volatile int sh_neuronOffsetTable[NUM_THREADS/INHIB_GROUPING_SZ+2];
		__shared__			int sh_firedTimeTable[NUM_THREADS/INHIB_GROUPING_SZ+2];
		__shared__			int sh_delayLength[NUM_THREADS/INHIB_GROUPING_SZ+2];
		__shared__			int sh_delayIndexStart[NUM_THREADS/INHIB_GROUPING_SZ+2];
		__shared__			int sh_timingInhib;

		const int grpId				= threadIdx.x/INHIB_GROUPING_SZ;		
		const int numGrps           = blockDim.x/INHIB_GROUPING_SZ;
		const int threadIdGrp		= threadIdx.x%INHIB_GROUPING_SZ;

		__shared__ volatile int sh_blkErrCode;

		if(ENABLE_MORE_CHECK) {			
			if(threadIdx.x<=0) 
				sh_blkErrCode = 0;		
		}

		// load the time table for inhibitory neuron firing
		// KILLME: we assume that maximum delay for inhibition is only 1ms
		int computedNeurons = 0;
		if (0==threadIdx.x) {
			sh_timingInhib = timingTableI[cur_time+gpuInfo.D]; // ??? check check ???
		}
		__syncthreads();

		int kPos = sh_timingInhib + (blockIdx.x*numGrps);

		__syncthreads();

		// Do as long as we have some valid inhibitory neuron
		while((kPos >=0)&&(kPos < secIfireCnt)) {		
			int fPos = -1;

			// a group of threads loads the delay information
			if (threadIdx.x < numGrps) {
				sh_neuronOffsetTable[threadIdx.x] = -1;
				fPos = kPos + threadIdx.x;

				// find the neuron id and also delay information from fPos
				if((fPos>=0)&&(fPos < secIfireCnt)) {
					atomicAdd((int*)&sh_inhNeuronCnt,1);
					int id = gpuPtrs.IfiringTable[fPos];
					int tPos = (gpuInfo.D+1)*id;
					sh_neuronOffsetTable[threadIdx.x]= gpuPtrs.cumulativePost[id];
					sh_delayLength[threadIdx.x]      = gpuPtrs.postDelayInfo[tPos].delay_length;
					sh_delayIndexStart[threadIdx.x]  = gpuPtrs.postDelayInfo[tPos].delay_index_start;
				}
			}

			__syncthreads();

			// useful to measure the load balance for each block..
			if(threadIdx.x==0)  computedNeurons += sh_inhNeuronCnt;

			// no more fired neuron from table... we just break from loop
			if (sh_inhNeuronCnt==0)	break;

			__syncthreads();

			int offset = sh_neuronOffsetTable[grpId];

			if (threadIdx.x == 0) 
				sh_inhNeuronCnt = 0;

			if (offset>=0) {
				int delId=threadIdGrp;
				while(delId < sh_delayLength[grpId]) {
					int d   = (sh_delayIndexStart[grpId]+delId);

					// get the post synaptic information for specific delay
					int post_info = gpuPtrs.postSynapticIds[offset+d];

					// get neuron id
					int p_i = (post_info&POST_SYN_NEURON_MASK);

					// get synaptic id
					int wt_i = (post_info>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK;
					
					// get the actual position of the synapses and other variables...
					int pos_ns = gpuPtrs.cumulativePre[p_i] + wt_i;

					if(ENABLE_MORE_CHECK) {
						if(p_i >= gpuInfo.N)         sh_blkErrCode = ICURRENT_UPDATE_ERROR1;
						if(wt_i >= gpuPtrs.Npre[p_i]) sh_blkErrCode = ICURRENT_UPDATE_ERROR2;				
						if (sh_blkErrCode) break;
					}

					if(TESTING) int id = atomicAdd(&generatedSpikesI, 1);
					//senderIdI[id] = post_info;

					int* tmp_I_set_p = ((int*)((char*)gpuPtrs.I_set + (wt_i>>5) * gpuInfo.I_setPitch) + p_i);
					atomicOr(tmp_I_set_p, 1 << (wt_i%32));
					//atomicOr(&gpuPtrs.I_set[wt_i/32][p_i], (1<<(wt_i%32))); //uncoalesced access
					//gpu_I_set[wt_i/32][p_i] += (1<<(wt_i%32));

					delId += INHIB_GROUPING_SZ;
				} // while(current neurons inhibition is done)
			} // if(valid inhibitory neuron)	

			__syncthreads();

		   if(ENABLE_MORE_CHECK)
			 if(sh_blkErrCode)  break;		

			kPos = kPos + (gridDim.x*numGrps);
		}

		MEASURE_GPU(4, computedNeurons);

		if(ENABLE_MORE_CHECK) {	
		   if (sh_blkErrCode) {
			     retErrCode = sh_blkErrCode;
			     return;
		   }
		}

	}

	/**********************************************************************************************/
	// helper functions..
	// check what is the errors...that has happened during
	// the previous iteration. This part is useful in debug or test mode
	// the simulator kernel sets appropriate errors  and returns some important values in the 
	// array "retErrVal".
	/**********************************************************************************************/
	int checkCudaErrors(int numBlocks) 
	{
	   int errCode = NO_KERNEL_ERRORS;

	   if(!ENABLE_MORE_CHECK) 
		   return errCode;

	   void* devPtr;
	   int errVal[MAX_NUM_BLOCKS][10];
	   
	   cudaThreadSynchronize();
	   cudaGetSymbolAddress(&devPtr, retErrCode);
	   cutilSafeCall( cudaMemcpy(&errCode, devPtr, sizeof(int), cudaMemcpyDeviceToHost));

	   if (errCode != NO_KERNEL_ERRORS) {
		  printf("\n RETURN ERROR CODE = %x\n", errCode);

		  cudaGetSymbolAddress(&devPtr, "retErrVal");
		  cutilSafeCall( cudaMemcpy(&errVal, devPtr, sizeof(errVal), cudaMemcpyDeviceToHost));

		  for(int j=0; j < 1; j++) {
			  if (1) /*errVal[j][0]==0xdead)*/ {
				  fprintf(stderr, "Err code = %x, total err val is %d\n", errVal[j][0], errVal[j][1]);
				  for(int i=0; i < (errVal[j][1]+2); i++) {
					  fprintf(stderr, "ErrVal[%d][%d] = %d\n", j, i, errVal[j][i]);	   
					  getchar();
				  }
			  }
		  }
	   }	

	   fflush(stderr);
	   fflush(stdout);
	   return errCode;
	}
	
	void CpuSNN::copyConnections(network_ptr_t* dest,int kind, int allocateMem)
	{
		void* devPtr;

		// allocateMem memory only if destination memory is not allocated !!!
		assert(allocateMem && (dest->allocated != 1));
		
		if(kind==cudaMemcpyHostToDevice) {
			assert(dest->memType == GPU_MODE);
		} 
		else {
			assert(dest->memType == CPU_MODE);
		}
		
		net_Info.I_setLength = ((int)((3*M-1)/32));
		assert(net_Info.I_setLength >= 0);
		if(allocateMem) 
			cudaMallocPitch( (void**) &dest->I_set, &net_Info.I_setPitch, sizeof(int)*N, net_Info.I_setLength);
		assert(net_Info.I_setPitch >= 0);
		cutilSafeCall( cudaMemset( dest->I_set, 0, net_Info.I_setPitch*net_Info.I_setLength));

		//connection synaptic lengths and cumulative lengths...
		if(allocateMem)   	cutilSafeCall( cudaMalloc((void**) &dest->Npre, sizeof(short)*N));
		cutilSafeCall( cudaMemcpy( dest->Npre, Npre, sizeof(short)*N, cudaMemcpyHostToDevice));

		//presyn excitatory connections
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->Npre_exc, sizeof(short)*N));
		cutilSafeCall( cudaMemcpy( dest->Npre_exc, Npre_exc, sizeof(short)*N, cudaMemcpyHostToDevice));

		//number of postsynaptic connections
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->Npost, sizeof(short)*N));
		cutilSafeCall( cudaMemcpy( dest->Npost, Npost, sizeof(short)*N, cudaMemcpyHostToDevice));

		//beginning position for the post-synaptic information
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->cumulativePost, sizeof(int)*N));
		cutilSafeCall( cudaMemcpy( dest->cumulativePost, cumulativePost, sizeof(int)*N, cudaMemcpyHostToDevice));

		//beginning position for the pre-synaptic information
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->cumulativePre, sizeof(int)*N));
		cutilSafeCall( cudaMemcpy( dest->cumulativePre, cumulativePre, sizeof(int)*N, cudaMemcpyHostToDevice));

		// actual post synaptic connection information...
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->postSynapticIds, sizeof(int)*(postSynCnt+10)));
		cutilSafeCall( cudaMemcpy( dest->postSynapticIds, postSynapticIds, sizeof(int)*(postSynCnt+10), cudaMemcpyHostToDevice));
		net_Info.postSynCnt = postSynCnt;

		// actual post synaptic delay metric
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->postDelayInfo, sizeof(int)*N*(D+1)));
		cutilSafeCall( cudaMemcpy( dest->postDelayInfo, postDelayInfo, sizeof(int)*N*(D+1), cudaMemcpyHostToDevice));

		//neuron input current...
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->I, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->I, I, sizeof(float)*N, cudaMemcpyHostToDevice));

		// allocate randomPtr.... containing the firing information for the random firing neurons
		assert(numRandNeurons >= 0);
		if(allocateMem)  cutilSafeCall( cudaMalloc( (void**) &dest->randId, sizeof(int)*numRandNeurons));
		net_Info.numRandNeurons=numRandNeurons;
		
		//copy the properties of the noise generator here..
		if(allocateMem)  cutilSafeCall( cudaMalloc( (void**) &dest->noiseGenProp, sizeof(noiseGenProperty_t)*numNoise));
		cutilSafeCall( cudaMemcpy( dest->noiseGenProp, &noiseGenGroup[0], sizeof(noiseGenProperty_t)*numNoise, cudaMemcpyHostToDevice));
	}

	void CpuSNN::generateHistogram_GPU(int* sHist, int* sdHist, int histSize)
	{
		int blkSize  = 128;
		int gridSize = 64;		

		assert(histSize == HISTOGRAM_SIZE);
		void* dev_sHist;
		void* dev_sdHist;

		//clear and reset gpu memory histogram array
        cutilSafeCall( cudaGetSymbolAddress(&dev_sHist,  "gpu_sHist"));
		cutilSafeCall( cudaGetSymbolAddress(&dev_sdHist, "gpu_sdHist"));
		cutilSafeCall( cudaMemset(dev_sHist,  0, sizeof(int)*(histSize+2)));
		cutilSafeCall( cudaMemset(dev_sdHist, 0, sizeof(int)*(histSize+2)));

		kernel_getHistogram <<<gridSize, blkSize>>> ();
        cutilCheckMsg("generateHistogram kernel failed\n");
		int errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);

		//copy back the new histogram array from gpu memory to cpu memory
        cutilSafeCall( cudaMemcpy(sHist,  dev_sHist, sizeof(int)*(histSize+2),  cudaMemcpyDeviceToHost));
		cutilSafeCall( cudaMemcpy(sdHist, dev_sdHist, sizeof(int)*(histSize+2), cudaMemcpyDeviceToHost));
	}
	
	// allocate necessary memory for the GPU...
	void CpuSNN::copyState(network_ptr_t* dest, int kind, int allocateMem)
	{
		assert(N != 0);
		assert(preSynLength !=0);

		//neuron information...
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->a, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->a, a, sizeof(float)*N, cudaMemcpyHostToDevice));
		
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->b, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->b, b, sizeof(float)*N, cudaMemcpyHostToDevice));		

		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->c, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->c, c, sizeof(float)*N, cudaMemcpyHostToDevice));
		
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->d, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->d, d, sizeof(float)*N, cudaMemcpyHostToDevice));
		
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->u, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->u, u, sizeof(float)*N, cudaMemcpyHostToDevice));

		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->v, sizeof(float)*N));
		cutilSafeCall( cudaMemcpy( dest->v, v, sizeof(float)*N, cudaMemcpyHostToDevice));

		// neuron firing time..
		if(allocateMem)     cutilSafeCall( cudaMalloc( (void**) &dest->lastSpikeTime, sizeof(int)*N));
		cutilSafeCall( cudaMemcpy( dest->lastSpikeTime, lastSpikeTime, sizeof(int)*N, cudaMemcpyHostToDevice));

		// synaptic information based
		if(allocateMem)		cutilSafeCall( cudaMalloc( (void**) &dest->wt, sizeof(float)*preSynLength));
		cutilSafeCall( cudaMemcpy( dest->wt, wt, sizeof(float)*preSynLength, cudaMemcpyHostToDevice));
		
		// synaptic weight derivative
		if(allocateMem)		cutilSafeCall( cudaMalloc( (void**) &dest->wtChange, sizeof(float)*preSynLength));
		cutilSafeCall( cudaMemcpy( dest->wtChange, wtChange, sizeof(float)*preSynLength, cudaMemcpyHostToDevice));

		// firing time for individual synapses
		if(allocateMem)		cutilSafeCall( cudaMalloc( (void**) &dest->synSpikeTime, sizeof(int)*preSynLength));
		cutilSafeCall( cudaMemcpy( dest->synSpikeTime, synSpikeTime, sizeof(int)*preSynLength, cudaMemcpyHostToDevice));
		net_Info.preSynLength = preSynLength;

		assert(net_Info.maxSpikesInhibitory != 0);
		if(allocateMem) {
			assert(dest->IfiringTable == NULL);
			assert(dest->EfiringTable == NULL);
		}
		// allocate inhibitory neuron'wt firing table
		if(allocateMem)		cutilSafeCall( cudaMalloc( (void**) &dest->IfiringTable, sizeof(int)*net_Info.maxSpikesInhibitory));
		cutilSafeCall( cudaMemcpy( dest->IfiringTable, IfiringTable, sizeof(int)*net_Info.maxSpikesInhibitory, cudaMemcpyHostToDevice));

		// allocate excitatory neuron'wt firing table
		if(allocateMem)		cutilSafeCall( cudaMalloc( (void**) &dest->EfiringTable, sizeof(int)*net_Info.maxSpikesExcitatory));
		cutilSafeCall( cudaMemcpy( dest->EfiringTable, EfiringTable, sizeof(int)*net_Info.maxSpikesExcitatory, cudaMemcpyHostToDevice));
	}

	//Initialize the thalamic input to network
	void CpuSNN::initThalInput_GPU(int with_stdp)
	{
		int blkSize  = 128;
		int gridSize = 64;
		// copy the generated random number into the randId array
		cutilSafeCall( cudaMemcpy(gpu_Ptrs.randId, randNeuronId, sizeof(int)*numRandNeurons, cudaMemcpyHostToDevice));

		// initialze the thalamic input current in to the network
		kernel_initThalInput <<<gridSize,blkSize >>> ();		
		cutilCheckMsg("initThalInput kernel failed\n");
		int errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);
	}
	
	void CpuSNN::findFiring_GPU(int with_stdp)
	{
	    int blkSize  = 128;
		int gridSize = 64;
		kernel_findFiring <<<gridSize,blkSize >>> (simTimeMs, simTimeSec, simTime, with_stdp);
		cutilCheckMsg("findFiring kernel failed\n");
		int errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);		
	}
	
	void CpuSNN::updateTimingTable_GPU(int with_stdp)
	{
		int blkSize  = 128;
		int gridSize = 64;
        gpu_timingTableUpdate <<<gridSize,blkSize >>> (simTimeMs);
		cutilCheckMsg("timing Table update kernel failed\n");
		int errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);	
	}
	
	void CpuSNN::doCurrentUpdate_GPU(int with_stdp)
	{
		int blkSize  = 128;
		int gridSize = 64;

		if(M>=550)
			;//gpu_doCurrentUpdateExcitLarge <<<gSize, bSize>>>(simTimeMs,simTimeSec,simTime,with_stdp);
		else
			gpu_doCurrentUpdateExcit <<<gridSize, blkSize>>>(simTimeMs,simTimeSec,simTime,with_stdp);
		
		CUT_CHECK_ERROR("Kernel execution failed");	
		int errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);		

		gpu_doCurrentUpdateInhib <<<gridSize, blkSize>>>(simTimeMs,simTimeSec,simTime);
		CUT_CHECK_ERROR("Kernel execution failed");	
		errCode = checkCudaErrors(gridSize);
		assert(errCode == NO_KERNEL_ERRORS);
	}

	void CpuSNN::initGPU()
	{
		int gridSize = 1;
		int blkSize  = 128;		
		kernel_init <<< 1, 128 >>> ();
		cutilCheckMsg("initGPU kernel failed\n");
		int errCode = checkCudaErrors(gridSize);	
		assert(errCode == NO_KERNEL_ERRORS);
		checkInitialization();
	}

	void CpuSNN::checkInitialization()
	{
		int testTable[10];
		// we write some number in this table..
		// write that to the GPU memory. Read it back
		// using the texture access and check if everything is correctly initialized
		testTable[0] = 11; testTable[1] = 12; testTable[2] = 13;
		testTable[3] = 14; testTable[4] = 15; testTable[5] = 16;
		cutilSafeCall( cudaMemcpyToSymbol("timingTableE", testTable, sizeof(int)*(10), 0, cudaMemcpyHostToDevice));
		check_GPU_init <<< 1, 128 >>>();
		cutilCheckMsg("check GPU failed\n");

		// read back the intialization and ensure that they are okay
		fprintf(fpLog, "Checking initialization of GPU...\n");
		void *devPtr;
		float errVal[MAX_NUM_BLOCKS][10];
		cudaGetSymbolAddress(&devPtr, retErrVal);
		cutilSafeCall( cudaMemcpy(&errVal, devPtr, sizeof(errVal), cudaMemcpyDeviceToHost));
		for(int i=0; i < 2; i++) {
			for(int j=0; j < (int)errVal[i][0]; j++) {
				fprintf(fpLog, "val[%d] = %f\n", j, errVal[i][j]);
			}			  
		}
		fprintf(fpLog, "Checking done...\n");
		cutilSafeCall( cudaMemcpyToSymbol("timingTableE", EtimeTable, sizeof(int)*(10), 0, cudaMemcpyHostToDevice));
	}

	void testSpikeSenderReceiver(FILE* fpLog)
	{
		if(TESTING) {
			int sendId[1000];
			int recId[1000];
			int genI, genE, recE, recI;
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &genE, "generatedSpikesE", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &recE, "receivedSpikesE", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &genI, "generatedSpikesI", sizeof(int), 0, cudaMemcpyDeviceToHost));
			CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &recI, "receivedSpikesI", sizeof(int), 0, cudaMemcpyDeviceToHost));

			//fprintf(stderr, "generatedE = %d, receivedE = %d generatedI = %d, receivedI = %d\n", genE, recE, genI, recI);
			fprintf(fpLog, "generatedE = %d, receivedE = %d generatedI = %d, receivedI = %d\n", genE, recE, genI, recI);

			assert(genE == recE);
			assert(genI == recI);
/*
			genI = 0; genE = 0; recI = 0; recE = 0;
			CUDA_SAFE_CALL( cudaMemcpyToSymbol( "generatedSpikesE", &genE, sizeof(int), 0, cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL( cudaMemcpyToSymbol( "receivedSpikesE", &recE, sizeof(int), 0, cudaMemcpyHostToDevice));	
			CUDA_SAFE_CALL( cudaMemcpyToSymbol( "generatedSpikesI", &genI, sizeof(int), 0, cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL( cudaMemcpyToSymbol( "receivedSpikesI", &recI, sizeof(int), 0, cudaMemcpyHostToDevice));	
			*/
		}

		/*
		CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &sendId, "senderIdE", sizeof(int)*genE, 0, cudaMemcpyDeviceToHost));
		CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &sendId[genE], "senderIdI", sizeof(int)*genI, 0, cudaMemcpyDeviceToHost));
		CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &recId, "receiverId", sizeof(int)*(recI+recE), 0, cudaMemcpyDeviceToHost));

		int cnt=0;
		for(int i = 0; i < genE+genI; i++) {
			int found = false;
			for(int j=0; j < 0; j++) {
				if(recId[j] == (sendId[i])) {
					recId[j]= -1;
					cnt++;
					found = false;
				}
			}
			if(!found)
			   fprintf(fpLog, "sendE[%d] = %d, syn = %d\n", i, sendId[i]&POST_SYN_NEURON_MASK
			                                                 , ((sendId[i]>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK));
		}
		if(cnt != (recI+recE)) {
			for(int i = 0; i < recE+recI; i++) {
				if(recId[i]!=-1) {
					fprintf(fpLog, "extra[%d] = %d, syn = %d\n", i, 
						  recId[i]&POST_SYN_NEURON_MASK,
						((recId[i]>>POST_SYN_NEURON_BITS)&POST_SYN_CONN_MASK));
				}
			}
		}
		*/

	}


	void CpuSNN::globalStateUpdate_GPU(int with_stdp)
	{
		int blkSize  = 128;
		int gridSize = 64;

		gpu_globalStateUpdate <<<gridSize, blkSize>>>();
		CUT_CHECK_ERROR("Kernel execution failed");	
		int errCode = checkCudaErrors(gridSize);  		
		assert(errCode == NO_KERNEL_ERRORS);			
	}	

	void CpuSNN::doGPUSim(int with_stdp)
	{
		initThalInput_GPU(with_stdp);		
		findFiring_GPU(with_stdp);		
		updateTimingTable_GPU(with_stdp);
		doCurrentUpdate_GPU(with_stdp);
		globalStateUpdate_GPU(with_stdp);
		testSpikeSenderReceiver(fpLog);
		//showStatus(GPU_MODE,1, fpLog);
		//simTime++;
	}
	
	void CpuSNN::updateStateAndFiringTable_GPU()
	{			
		int blkSize  = 128;
		int gridSize = 64;

		kernel_updateWeightsFiring  <<<gridSize, blkSize>>> ();
		kernel_updateWeightsFiring2 <<<gridSize, blkSize>>> ();			
	}
	
	void CpuSNN::copyFiringInfo_GPU()
	{
		int gpu_secIfireCnt, gpu_secEfireCnt;
		CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &gpu_secEfireCnt, "secEfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
		CUDA_SAFE_CALL( cudaMemcpyFromSymbol( &gpu_secIfireCnt, "secIfireCnt", sizeof(int), 0, cudaMemcpyDeviceToHost));
		spikeCountAll1sec = gpu_secIfireCnt + gpu_secEfireCnt;
		spikeCountInhibitory1sec  = gpu_secIfireCnt;
        cutilSafeCall( cudaMemcpy(EfiringTable, gpu_Ptrs.EfiringTable, sizeof(int)*gpu_secEfireCnt, cudaMemcpyDeviceToHost));
	    cutilSafeCall( cudaMemcpy(IfiringTable, gpu_Ptrs.IfiringTable, sizeof(int)*gpu_secIfireCnt, cudaMemcpyDeviceToHost));
		cutilSafeCall( cudaMemcpyFromSymbol(EtimeTable, "timingTableE", sizeof(int)*(1000+D+1), 0, cudaMemcpyDeviceToHost));
		cutilSafeCall( cudaMemcpyFromSymbol(ItimeTable, "timingTableI", sizeof(int)*(1000+D+1), 0, cudaMemcpyDeviceToHost));
	}

	// Allocates required memory and then initialize the GPU
	void CpuSNN::allocate()
	{
		//ensure that we dont do all the above optimizations again		
		assert(doneReorganization == true);		

		net_Info.N  = N;
		net_Info.M  = M;
		net_Info.D  = D;
		net_Info.Ne = Ne;
		net_Info.Ni	= Ni;		
		net_Info.numNoise  = numNoise;
		net_Info.ALPHA_LTP = ALPHA_LTP;
		net_Info.TAU_LTP   = TAU_LTP;
		net_Info.ALPHA_LTD = ALPHA_LTD;
		net_Info.TAU_LTD   = TAU_LTD;
		net_Info.maxSpikesExcitatory = maxSpikesExcitatory;
		net_Info.maxSpikesInhibitory = maxSpikesInhibitory;	
		net_Info.maxWeight    = maxWeight;
		net_Info.maxWeightChange   = maxWeightChange;
		net_Info.STDPenabled = STDPenabled;

		gpu_Ptrs.memType = GPU_MODE;
		int dev = cutGetMaxGflopsDeviceId();
		fprintf(stdout, "Device with maximum GFLOPs is : %d\n", dev);
		cudaDeviceProp deviceProp;
		cutilSafeCall(cudaGetDeviceProperties(&deviceProp, dev));
		fprintf(stdout, "\nDevice %d: \"%s\"\n", dev, deviceProp.name);
		if (deviceProp.major == 1 && deviceProp.minor < 3) {
			printf("GPU SNN does not support NVidia cards older than version 1.3\n");
			exit(1);
		}
		assert(deviceProp.major >= 1);
		assert(deviceProp.minor >= 3);		
		cudaThreadExit();
		cudaSetDevice( dev );
		cutilCheckMsg("cudaSetDevice failed\n");	

		// this table is useful for quick evaluation of the position of fired neuron
		// given a sequence of bits denoting the firing..
		initTableQuickSynId();
		copyConnections(&gpu_Ptrs,  cudaMemcpyHostToDevice, 1);
		copyState(&gpu_Ptrs, cudaMemcpyHostToDevice, 1);
				
		gpu_Ptrs.allocated = true;

		void* devPtr;
		// copy relevant pointers and network information to GPU
		cutilSafeCall( cudaMemcpyToSymbol("gpuPtrs", &gpu_Ptrs, sizeof(network_ptr_t)));
		cutilSafeCall( cudaMemcpyToSymbol("gpuInfo", &net_Info, sizeof(network_info_t)));

		// map the timing table to texture.. saves a lot of headache in using shared memory
		CUDA_SAFE_CALL ( cudaGetSymbolAddress(&devPtr, "timingTableE"));
		CUDA_SAFE_CALL ( cudaBindTexture(0, timingTableE_tex, devPtr, sizeof(int)*ROUNDED_TIMING_COUNT));
		CUDA_SAFE_CALL ( cudaGetSymbolAddress(&devPtr, "timingTableI"));
		CUDA_SAFE_CALL ( cudaBindTexture(0, timingTableI_tex, devPtr, sizeof(int)*ROUNDED_TIMING_COUNT));

		cutilSafeCall( cudaMemset( gpu_Ptrs.I, 0, sizeof(float)*N));

		initGPU();

	}
